import random
from typing import Union, Optional, Generator, List

from pydantic import BaseModel, ConfigDict, Field

import logging

logger = logging.getLogger(__name__)

from simpy import events

from prodsys.models import product_data

from prodsys.simulation import (
    process_models,
    request,
    auxiliary,
    process,
    router,
    resources,
    sim,
    sink,
    store,
    source,
    node,
)
from prodsys.simulation.state import StateTypeEnum, StateEnum
from prodsys.simulation.process import PROCESS_UNION, ReworkProcess


class ProductInfo(BaseModel):
    """
    Class that represents information of the current state of a product.

    Args:
        resource_ID (str): ID of the resource that the product is currently at.
        state_ID (str): ID of the state that the product is currently at.
        event_time (float): Time of the event.
        activity (state.StateEnum): Activity of the product.
        product_ID (str): ID of the product.
        state_type (state.StateTypeEnum): Type of the state.
    """

    resource_ID: str = Field(init=False, default=None)
    state_ID: str = Field(init=False, default=None)
    event_time: float = Field(init=False, default=None)
    activity: StateEnum = Field(init=False, default=None)
    product_ID: str = Field(init=False, default=None)
    state_type: StateTypeEnum = Field(init=False, default=None)

    model_config = ConfigDict(extra="allow")

    def log_finish_product(
        self,
        resource: Union[resources.Resource, sink.Sink, source.Source],
        _product: "Product",
        event_time: float,
    ):
        """
        Logs the finish of a product.

        Args:
            resource (Union[resources.Resource, sink.Sink, source.Source]): New resource of the product.
            _product (Product): Product that is finished.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.finished_product
        self.state_type = StateTypeEnum.sink

    def log_create_product(
        self,
        resource: Union[resources.Resource, sink.Sink, source.Source],
        _product: "Product",
        event_time: float,
    ) -> None:
        """
        Logs the creation of a product.

        Args:
            resource (Union[resources.Resource, sink.Sink, source.Source]): New resource of the product.
            _product (Product): Product that is created.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.created_product
        self.state_type = StateTypeEnum.source

    def log_start_process(
        self,
        resource: resources.Resource,
        _product: "Product",
        event_time: float,
        state_type: StateTypeEnum,
    ) -> None:
        """
        Logs the start of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.start_state
        self.state_type = state_type

    def log_end_process(
        self,
        resource: resources.Resource,
        _product: "Product",
        event_time: float,
        state_type: StateTypeEnum,
    ) -> None:
        """
        Logs the end of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.end_state
        self.state_type = state_type


Locatable = Union[resources.Resource, node.Node, source.Source, sink.Sink, store.Store]


class Product(BaseModel):
    """
    Class that represents a product in the discrete event simulation. For easier instantion of the class, use the ProductFactory at prodsys.factories.product_factory.

    Args:
        env (sim.Environment): prodsys simulation environment.
        product_data (product_data.ProductData): Product data that represents the meta information of the simulation product object.
        process_model (proces_models.ProcessModel): Process model that represents the required manufacturing processes and the current state of the product.
        transport_process (process.Process): Transport process that represents the required transport processes.
        product_router (router.Router): Router that is used to route the product object.
    """

    env: sim.Environment
    product_data: product_data.ProductData
    process_model: process_models.ProcessModel
    transport_process: Union[
        process.TransportProcess,
        process.RequiredCapabilityProcess,
        process.LinkTransportProcess,
    ]
    product_router: router.Router

    next_prodution_process: Optional[process.PROCESS_UNION] = Field(
        default=None, init=False
    )
    processes_needing_rework: List[process.Process] = Field(
        default_factory=list, init=False
    )
    blocking_rework_process_mappings: list[
        list[process.PROCESS_UNION, list[ReworkProcess]]
    ] = Field(default_factory=list, init=False)
    non_blocking_rework_process_mappings: list[
        list[process.PROCESS_UNION, list[ReworkProcess]]
    ] = Field(default_factory=list, init=False)
    process: events.Process = Field(default=None, init=False)
    current_locatable: Locatable = Field(default=None, init=False)
    finished_process: events.Event = Field(default=None, init=False)
    product_info: ProductInfo = Field(default_factory=ProductInfo, init=False)
    executed_production_processes: List = Field(default_factory=list, init=False)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def update_location(self, locatable: Locatable):
        """
        Updates the location of the product object.

        Args:
            locatable (Locatable): Locatable objects where product object currently is.
        """
        self.current_locatable = locatable
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "resource": self.current_locatable.data.ID,
                "event": f"Updated location to {self.current_locatable.data.ID}",
            }
        )

    def process_product(self):
        self.finished_process = events.Event(self.env)
        self.product_info.log_create_product(
            resource=self.current_locatable, _product=self, event_time=self.env.now
        )
        """
        Processes the product object in a simpy process. The product object is processed after creation until all required production processes are performed and it reaches a sink.
        """
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "event": f"Start processing of product",
            }
        )
        self.set_next_production_process()

        if self.product_data.auxiliaries:
            while True:
                auxiliary_request: request.AuxiliaryRequest = yield self.env.process(
                    self.product_router.route_auxiliary_to_product(self)
                )
                if not auxiliary_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "resource": auxiliary_request.resource.data.ID,
                    "aux": auxiliary_request.auxiliary.product_data.ID,
                    "process": auxiliary_request.process.process_data.ID,
                    "event": f"auxiliary request for {auxiliary_request.product.product_data.ID}",
                }
            )
            while True:
                auxiliary_transport_request: request.TransportResquest = (
                    yield self.env.process(
                        self.product_router.route_transport_resource_for_item(
                            auxiliary_request
                        )
                    )
                )
                if not auxiliary_transport_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "resource": auxiliary_transport_request.resource.data.ID,
                    "aux": auxiliary_transport_request.product.product_data.ID,
                    "process": auxiliary_transport_request.process.process_data.ID,
                    "origin": auxiliary_transport_request.origin.data.ID,
                    "target": auxiliary_transport_request.target.data.ID,
                    "event": f"starting auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}",
                }
            )
            yield self.env.process(
                auxiliary_request.auxiliary.request_process(auxiliary_transport_request)
            )
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "resource": auxiliary_transport_request.resource.data.ID,
                    "aux": auxiliary_transport_request.product.product_data.ID,
                    "process": auxiliary_transport_request.process.process_data.ID,
                    "origin": auxiliary_transport_request.origin.data.ID,
                    "target": auxiliary_transport_request.target.data.ID,
                    "event": f"finished waiting for auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}",
                }
            )

        while self.next_prodution_process:
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "process": self.next_prodution_process.process_data.ID,
                    "event": f"Start process of product",
                }
            )
            while True:
                production_request = yield self.env.process(
                    self.product_router.route_product_to_production_resource(self)
                )
                if not production_request:
                    yield self.env.timeout(0)
                    continue
                break
            while True:
                transport_request = yield self.env.process(
                    self.product_router.route_transport_resource_for_item(
                        production_request
                    )
                )
                if not transport_request:
                    yield self.env.timeout(0)
                    continue
                break
            yield self.env.process(self.request_process(transport_request))
            yield self.env.process(self.request_process(production_request))
            store_product = self.product_router.check_store_product(self)
            if store_product:
                logger.debug(
                    {
                        "ID": self.product_data.ID,
                        "sim_time": self.env.now,
                        "event": f"Store product in storage",
                    }
                )
                while True:
                    transport_to_storage_request = yield self.env.process(
                        self.product_router.route_product_to_storage(self)
                    )
                    if not transport_to_storage_request:
                        yield self.env.timeout(0)
                        continue
                    break
                yield self.env.process(
                    self.request_process(transport_to_storage_request)
                )
                logger.debug(
                    {
                        "ID": self.product_data.ID,
                        "sim_time": self.env.now,
                        "event": f"Product transported to storage",
                    }
                )

            self.set_next_production_process()
        while True:
            transport_to_sink_request = yield self.env.process(
                self.product_router.route_product_to_sink(self)
            )
            if not transport_to_sink_request:
                yield self.env.timeout(0)
                continue
            break
        yield self.env.process(self.request_process(transport_to_sink_request))
        self.product_info.log_finish_product(
            resource=self.current_locatable, _product=self, event_time=self.env.now
        )
        self.current_locatable.register_finished_product(self)
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "event": f"Finished processing of product",
            }
        )

        if self.product_data.auxiliaries:
            auxiliary_request.auxiliary.update_location(self.current_locatable)
            while True:
                auxiliary_transport_request: request.TransportResquest = (
                    yield self.env.process(
                        self.product_router.route_auxiliary_to_store(
                            auxiliary_request.auxiliary
                        )
                    )
                )
                if not auxiliary_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "resource": auxiliary_transport_request.resource.data.ID,
                    "aux": auxiliary_transport_request.product.product_data.ID,
                    "process": auxiliary_transport_request.process.process_data.ID,
                    "origin": auxiliary_transport_request.origin.data.ID,
                    "target": auxiliary_transport_request.target.data.ID,
                    "event": f"starting auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}",
                }
            )
            yield self.env.process(
                auxiliary_request.auxiliary.request_process(auxiliary_transport_request)
            )
            auxiliary_request.auxiliary.release_auxiliary_from_product()

    def request_process(self, processing_request: request.Request) -> Generator:
        """
        Requests the next production process of the product object from the next production resource by creating a request event and registering it at the environment.
        """
        if isinstance(processing_request, request.TransportResquest):
            type_ = StateTypeEnum.transport
        else:
            type_ = StateTypeEnum.production
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "resource": processing_request.resource.data.ID,
                "event": f"Request process {processing_request.process.process_data.ID} for {type_}",
            }
        )
        self.env.request_process_of_resource(request=processing_request)
        yield self.finished_process
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "resource": processing_request.resource.data.ID,
                "event": f"Finished process {processing_request.process.process_data.ID} for {type_}",
            }
        )
        self.product_info.log_end_process(
            resource=processing_request.resource,
            _product=self,
            event_time=self.env.now,
            state_type=type_,
        )
        self.finished_process = events.Event(self.env)
        if isinstance(processing_request.process, process.ReworkProcess):
            self.register_rework(processing_request.process)

    def add_needed_rework(self, failed_process: PROCESS_UNION) -> None:
        """
        Adds a process to the list of processes that need rework.

        Args:
            failed_process (PROCESS_UNION): Process that needs rework.
        """
        self.processes_needing_rework.append(failed_process)
        possible_rework_processes = self.product_router.get_rework_processes(
            self, failed_process
        )
        if not possible_rework_processes:
            raise ValueError(
                f"No rework processes found for process {failed_process.process_data.ID}"
            )
        if any(rework_process.blocking for rework_process in possible_rework_processes):
            self.blocking_rework_process_mappings.append(
                [failed_process, possible_rework_processes]
            )
        else:
            self.non_blocking_rework_process_mappings.append(
                [failed_process, possible_rework_processes]
            )

    def register_rework(self, rework_process: ReworkProcess) -> None:
        """
        Register a rework process that has been executed.
        """
        for process_rework_mapping in (
            self.blocking_rework_process_mappings
            + self.non_blocking_rework_process_mappings
        ):
            if not rework_process in process_rework_mapping[1]:
                continue
            reworked_process = process_rework_mapping[0]
            break
        self.processes_needing_rework.remove(reworked_process)
        in_blocking_list = any(
            reworked_process == process_rework_mapping[0]
            for process_rework_mapping in self.blocking_rework_process_mappings
        )
        if in_blocking_list:
            mapping_to_adjust = self.blocking_rework_process_mappings
        else:
            mapping_to_adjust = self.non_blocking_rework_process_mappings
        for index, process_rework_mapping in enumerate(mapping_to_adjust):
            if reworked_process == process_rework_mapping[0]:
                index_to_remove = index
                break
        mapping_to_adjust.pop(index_to_remove)

    def set_next_production_process(self):
        """
        Sets the next process of the product object based on the current state of the product and its process model.
        """
        # check if rework is needed due to blocking rework processes. If so, execute these rework processes
        if self.blocking_rework_process_mappings:
            process_mapping_to_rework = random.choice(
                self.blocking_rework_process_mappings
            )
            next_possible_processes = process_mapping_to_rework[1]
            self.next_prodution_process = random.choice(next_possible_processes)
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "event": f"Blocking rework as next process with ID {self.next_prodution_process.process_data.ID}",
                }
            )
            return

        next_possible_processes = self.process_model.get_next_possible_processes()

        # if all normal processes are done, i.e. the product has finished its process sequence, execute rework processes without blocking.
        if not next_possible_processes and self.non_blocking_rework_process_mappings:
            process_mapping_to_rework = random.choice(
                self.non_blocking_rework_process_mappings
            )
            failed_process = process_mapping_to_rework[0]
            next_possible_processes = process_mapping_to_rework[1]
            self.next_prodution_process = random.choice(next_possible_processes)
            logger.debug(
                {
                    "ID": self.product_data.ID,
                    "sim_time": self.env.now,
                    "event": f"Non-blocking rework as next process with ID {self.next_prodution_process.process_data.ID}",
                }
            )
            return
        # if all normal processes are done and no rework processes are needed, the product has finished its process sequence.
        if not next_possible_processes:
            self.next_prodution_process = None
            return

        self.next_prodution_process = random.choice(next_possible_processes)  # type: ignore
        self.process_model.update_marking_from_transition(self.next_prodution_process)  # type: ignore
        self.executed_production_processes.append(
            self.next_prodution_process.process_data.ID
        )
        logger.debug(
            {
                "ID": self.product_data.ID,
                "sim_time": self.env.now,
                "event": f"Next process {self.next_prodution_process.process_data.ID}",
            }
        )



from __future__ import annotations

from typing import Dict, List, TYPE_CHECKING, Optional, Tuple, Union

import numpy as np

from pathfinding.core.graph import Graph, GraphNode
from pathfinding.finder.dijkstra import DijkstraFinder


if TYPE_CHECKING:
    from prodsys.simulation import request, process, resources
    from prodsys.simulation.product import Locatable


def find_route(
    request: request.TransportResquest,
    process: process.LinkTransportProcess,
    find_route_to_origin: bool = False,
) -> List[Locatable]:
    """
    Finds the route for a transportation request.

    Args:
        request (request.TransportResquest): The transportation request.
        process (process.LinkTransportProcess): The process to find the route for.
        find_route_to_origin (bool, optional): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False). Defaults to False. This also indicates whether material is transported or not.

    Returns:
        List[Locatable]: The route as a list of locatable objects.
    """
    if request.route and not find_route_to_origin:
        return request.route
    route_finder = RouteFinder()
    route = route_finder.find_route(
        request=request, process=process, find_route_to_origin=find_route_to_origin
    )
    if route and not find_route_to_origin:
        request.set_route(route=route)
    return route


class RouteFinder:
    """
    Class representing a route finder for transportation requests in a graph.
    """

    def __init__(self):
        """
        Initializes two empty lists of nodes and node locations.
        """
        self.nodes: Dict[str, GraphNode] = {}

    def find_route(
        self,
        request: request.TransportResquest,
        process: process.LinkTransportProcess,
        find_route_to_origin: bool = False,
    ) -> List[Locatable]:
        """
        The general function which includes all sub functions to find the shortest route for a TransportRequest.

        Args:
            request (TransportResquest): The transportation request.
            process (LinkTransportProcess): The process to find the route for.
            find_route_to_origin (bool): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False).

        Returns:
            List[Locatable]: The route as a list of locatable objects.
        """
        origin, target = self.get_route_origin_and_target_locatables(
            request=request, route_to_origin=find_route_to_origin
        )
        edges = self.process_links_to_graph_edges(
            links=process.links,
            origin=origin,
            target=target,
            empty_transport=find_route_to_origin,
        )  # finding route to origin is always an empty transport, since material is picked up at the origin

        # TODO: also add functionality to make directional edges for conveyors, where backwards routing is not possibly.
        origin_node, target_node = self.get_route_origin_and_target(
            request=request, route_to_origin=find_route_to_origin
        )

        graph = Graph(edges=edges, bi_directional=True)
        if not origin_node or not target_node:
            return []
        graph_node_path = self.find_graphnode_path(origin_node, target_node, graph)
        if not graph_node_path:
            return []
        route = self.convert_node_path_to_locatable_route(
            graph_node_path=graph_node_path, links=process.links
        )
        from prodsys.simulation.resources import TransportResource

        if isinstance(origin, TransportResource):
            route = route[1:]
        return route

    def process_links_to_graph_edges(
        self,
        links: List[List[Locatable]],
        origin: Locatable,
        target: Locatable,
        empty_transport: bool,
    ) -> List[Tuple[GraphNode, GraphNode, int]]:
        """
        Processes the given links to create (Graph)-edges for the graph.

        Args:
            links (List[List[Locatable]]): The links as a list of lists of locatable objects.
            origin (Locatable): The origin locatable object.
            target (Locatable): The target locatable object.
            empty_transport (bool): Indicates whether the transport is empty or a product / auxiliary is transported.

        Returns:
            List[Tuple[Graphnode, Graphnode, int]]: The edges as a list of tuples (with a start_node, end_node and related costs).
        """
        # TODO: make the imports at top or bottom of file
        from prodsys.simulation.resources import ProductionResource, TransportResource
        from prodsys.simulation.store import Store

        pathfinder_edges = []

        if isinstance(origin, TransportResource):
            # this is necessary since a transport resource can be initialized with a random location, so a link is needed for the first drive
            origin_location = origin.get_location()
            closest_locatable = None
            for link in links:
                for locatable in link:
                    if closest_locatable is None:
                        closest_locatable = locatable
                    if self.calculate_cost(
                        origin_location, locatable.get_location()
                    ) < self.calculate_cost(
                        origin_location, closest_locatable.get_location()
                    ):
                        closest_locatable = locatable
            links.append([origin, closest_locatable])

        for link in links:
            link_origin = link[0]
            link_target = link[1]
            if (
                not isinstance(link_origin, (ProductionResource, Store))
                or not link_origin == origin
            ):
                origin_location = link_origin.get_location()
            elif empty_transport:
                origin_location = link_origin.get_input_location()
            elif not empty_transport:
                origin_location = link_origin.get_output_location()

            if (
                not isinstance(link_target, (ProductionResource, Store))
                or not link_target == target
            ):
                target_location = link_target.get_location()
            elif empty_transport:
                target_location = link_target.get_output_location()
            elif not empty_transport:
                target_location = link_target.get_input_location()

            edge = self.get_graph_edge(
                link_origin,
                link_target,
                origin_location=origin_location,
                target_location=target_location,
            )
            pathfinder_edges.append(edge)

        return pathfinder_edges

    def get_graph_edge(
        self,
        origin: Locatable,
        target: Locatable,
        origin_location: list[float],
        target_location: list[float],
    ) -> Tuple[GraphNode, GraphNode, int]:
        """
        Creates a graph edge for a link.

        Args:
            link (List[Locatable]): The link as a list of two locatable objects.
            origin_location (List[float]): The origin location.
            target_location (List[float]): The target location.

        Returns:
            Tuple[GraphNode, GraphNode, int]: A tuple containing the origin and target nodes for a link and the cost.
        """
        cost = self.calculate_cost(origin_location, target_location)
        origin_graph_node = self.get_graph_node_for_locatable(origin, origin_location)
        target_graph_node = self.get_graph_node_for_locatable(target, target_location)
        return origin_graph_node, target_graph_node, cost

    def get_existing_graph_node_for_locatable(
        self, locatable: Locatable
    ) -> Optional[GraphNode]:
        """
        Gets an existing graph node for a Locatable.

        Args:
            locatable (Locatable): The locatable object to get the graph node for.

        Returns:
            Optional[GraphNode, None]: The graph node or None.
        """
        if locatable.data.ID in self.nodes:
            return self.nodes[locatable.data.ID]
        return None

    def get_graph_node_for_locatable(
        self, locatable: Locatable, location: list[float]
    ) -> GraphNode:
        """
        Creates a graph node for a locatable.

        Args:
            locatable (Locatable): The locatable to create the graph node for.

        Returns:
            GraphNode: The graph node.
        """
        # TODO: make the imports at top or bottom of file
        from prodsys.simulation.resources import ProductionResource
        from prodsys.simulation.store import Store

        existing_node = self.get_existing_graph_node_for_locatable(locatable)
        if existing_node:
            return existing_node

        new_graph_node = GraphNode(node_id=locatable.data.ID)
        self.nodes[locatable.data.ID] = new_graph_node
        return new_graph_node

    def calculate_cost(self, node1: List[float], node2: List[float]) -> float:
        """
        Calculates the cost between two nodes based on the euclidean distance.

        Args:
            node1 (List[float]): The first node.
            node2 (List[float]): The second node.

        Returns:
            float: The cost.
        """
        # TODO: maybe use here the time model feature to weight the distance with the time it takes to travel it to optimize for fastest paths and not shortest paths
        return np.sqrt((node1[0] - node2[0]) ** 2 + (node1[1] - node2[1]) ** 2)

    def get_route_origin_and_target_locatables(
        self, request: request.TransportResquest, route_to_origin: bool
    ) -> Tuple[Optional[Locatable], Optional[Locatable]]:
        """
        Gets the origin and target locatable

        Args:
            request (request.TransportResquest): The transportation request.
            route_to_origin (bool): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False).

        Returns:
            Tuple[Optional[Locatable], Optional[Locatable]]: A tuple containing the origin and target locatable objects for the transport request.
        """
        if route_to_origin:
            origin_locatable = request.resource.controller._current_locatable
            target_locatable = request.origin
        else:
            origin_locatable = request.origin
            target_locatable = request.target
        return origin_locatable, target_locatable

    def get_route_origin_and_target(
        self, request: request.TransportResquest, route_to_origin: bool
    ) -> Tuple[Optional[GraphNode], Optional[GraphNode]]:
        """
        Converts the origin and target of the transport request to graph nodes.

        Args:
            request (TransportResquest): The transportation request.
            route_to_origin (bool): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False).

        Returns:
            Tuple[Optional[GraphNode], Optional[GraphNode]]: A tuple containing the origin and target graph nodes for the transport request and the route_to_origin flag.
        """
        (
            origin_locatable,
            target_locatable,
        ) = self.get_route_origin_and_target_locatables(
            request=request, route_to_origin=route_to_origin
        )
        origin_graph_node = self.get_existing_graph_node_for_locatable(origin_locatable)
        target_graph_node = self.get_existing_graph_node_for_locatable(target_locatable)
        return origin_graph_node, target_graph_node

    def find_graphnode_path(
        self, origin: GraphNode, target: GraphNode, graph: Graph
    ) -> List[GraphNode]:
        """
        Finds the path between the origin and target graph nodes.

        Args:
            origin (GraphNode): The origin node.
            target (GraphNode): The target node.
            graph (Graph): The graph.

        Returns:
            The path as a list of graph nodes.
        """
        finder = DijkstraFinder()
        path, _ = finder.find_path(origin, target, graph)
        if path is None:
            return []
        return path

    def convert_node_path_to_locatable_route(
        self, graph_node_path: List[GraphNode], links: List[List[Locatable]]
    ) -> List[Locatable]:
        """
        Converts the path of graph nodes to a route of Locatable objects.

        Args:
            graph_node_path (List[GraphNode]): The path as a list of graph nodes.
            links (List[List[Locatable]]): The given links list.

        Returns:
            List[Locatable]: The route as a list of locatable.
        """
        route = []
        seen_node_ids = []  # no node several times in the path

        for node in graph_node_path:
            for link in links:
                for locatable in link:
                    if (
                        node.node_id == locatable.data.ID
                        and locatable.data.ID not in seen_node_ids
                    ):
                        route.append(locatable)
                        seen_node_ids.append(locatable.data.ID)
        return route



"""
NodeData objects are used in prodsys to represent Locations in the production system that serve as nodes in a link. They are used in the 
`LinkTransportProcess` class to represent locations where routes can cross.
"""
from hashlib import md5
from typing import List

from pydantic import ConfigDict

from prodsys.models.core_asset import CoreAsset, Locatable


class NodeData(CoreAsset, Locatable):
    """
    Represents a node data object of a link for a Transport process.

    Attributes:
        ID (str): ID of the node.
        description (str): Description of the node.
        location (List[float]): Location of the node. It has to be a list of length 2.
    """

    def hash(self) -> str:
        """
        Hashes the node data object.

        Returns:
            str: Hash of the node data object.
        """
        return Locatable.hash(self)

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "N1",
                    "description": "Node 1",
                    "location": [0.0, 0.0],
                }
            ]
        }
    )



from typing import Optional, Union
from prodsys.express import core
from prodsys.models import queue_data


from pydantic import Field, conlist
from pydantic.dataclasses import dataclass


from uuid import uuid1


@dataclass
class Queue(core.ExpressObject):
    """
    Represents a storage object for products. A queue has no location, but is part of a resource / source / sink and has the same location as the resource / source / sink.

    The capacity of the queue is the maximum number of products that can be stored in the queue.

    Attributes:
            ID (Optional[str]): The unique identifier of the storage.
            capacity (Union[int, float]): The capacity of the storage.
    """

    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    capacity: Union[int, float] = 0

    def to_model(self) -> queue_data.QueueData:
        """
        Converts the storage object to a `StorageData` model.

        Returns:
                StorageData: The converted storage data.
        """
        return queue_data.QueueData(
            ID=self.ID,
            description="",
            capacity=self.capacity,
        )


@dataclass
class Store(Queue):
    """
    A store is a storage object for products / auiliaries. It has a location, an input location, and an output location. The input location is the location where products are stored, and the output location is the location where products are retrieved.

    A store can be understood as queue that is independent of a resource / source / sink concerning its location.

    Attributes:
            location (List[float]): The location coordinates of the storage.
            input_location (Optional[List[float]]): The input location coordinates of the storage.
            output_location (Optional[List[float]]): The output location coordinates of the storage.
            ID (Optional[str]): The unique identifier of the storage.
            capacity (Union[int, float]): The capacity of the storage.
    """

    location: list[float] = Field(..., min_length=2, max_length=2)
    input_location: Optional[list[float]] = Field(None, min_length=2, max_length=2)
    output_location: Optional[list[float]] = Field(None, min_length=2, max_length=2)
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    capacity: Union[int, float] = 0

    def to_model(self) -> queue_data.StoreData:
        """
        Converts the storage object to a `StorageData` model.

        Returns:
                StorageData: The converted storage data.
        """
        return queue_data.StoreData(
            ID=self.ID,
            description="",
            capacity=self.capacity,
            location=self.location,
            input_location=self.input_location,
            output_location=self.output_location,
        )



from typing import Annotated, Dict, Optional
from pydantic import BaseModel, Field

from prodsys.adapters.json_adapter import JsonProductionSystemAdapter


AdapterHash = Annotated[
    str, Field(..., description="MD5 hash identifier of the individual configuration")
]
AdapterId = Annotated[str, Field(..., description="ID of the adapter")]
GenerationIndex = Annotated[str, Field(..., description="Index of the generation")]

class SolutionMetadata(BaseModel):
    generation: GenerationIndex
    ID: AdapterId

OptimizationSolutionDict = dict[AdapterHash, SolutionMetadata]  # Dict of adapter ID to hash

class OptimizationProgress(BaseModel):
    total_steps: int = Field(0, description="Total number of steps in the optimization process")
    completed_steps: int = Field(0, description="Number of steps completed")
    hashes: dict[AdapterId, AdapterHash] = Field({}, description="Mapping of adapter ID to hash")

class OptimizationSolutions(BaseModel):
    current_generation: str = Field("0", description="Current generation index")
    hashes: OptimizationSolutionDict = Field({}, description="Mapping of adapter ID to hash")


class FitnessData(BaseModel):
    agg_fitness: float = Field(..., description="Aggregated fitness value")
    fitness: list[float] = Field(..., description="List of fitness components")
    objective_names: Optional[list[str]] = Field(None, description="List of objective names for the fitness components")
    time_stamp: float = Field(..., description="Timestamp of the fitness data in seconds since start of optimization")
    hash: AdapterHash
    production_system: Optional[JsonProductionSystemAdapter] = Field(None, description="Production system configuration")
    event_log_dict: Optional[dict] = Field(None, description="Event log dictionary")


FitnessEntry = Annotated[dict[AdapterId, FitnessData], Field(..., description="Mapping of adapter ID to its fitness data")]
OptimizationResults = Annotated[dict[GenerationIndex, FitnessEntry], Field(..., description="Mapping of generation index to its fitness data")]

def get_empty_optimization_results() -> OptimizationResults:
    return {"0": {}}



from hashlib import md5
from typing import Any, Optional
from pydantic import BaseModel, ConfigDict, Field, conlist, model_validator


class CoreAsset(BaseModel):
    """
    Class that represents a the core asset, consisting of an ID and a description.

    Args:
        ID (str): The unique ID of the core asset.
        description (str): The description of the core asset.
    """

    ID: str
    description: str

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "ID": "Example Core Asset",
                "description": "Asset data for Example Core Asset",
            }
        }
    )


class Locatable(BaseModel):
    """
    Class that represents a locatable entity, consisting of an ID, a description and a location.

    Args:
        location (list[float]): The location of the locatable asset. It has to be a list of length 2.
    """

    location: list[float] = Field(..., min_length=2, max_length=2)

    def hash(self) -> str:
        """
        Returns a unique hash for the locatable asset considering its location, input location and output location.

        Returns:
            str: Hash of the locatable asset.
        """
        return md5((str(self.location)).encode("utf-8")).hexdigest()


class InLocatable(Locatable):
    """
    Class that represents a locatable entity with an input location. The location parameter is the center point of the asset. It needs to be defined. The input location parameter specifies where products are added to the asset. If not specified, the location is set as input location.

    Args:
        location (list[float]): The location of the locatable asset. It has to be a list of length 2.
        input_location (list[float], optional): The input location of the locatable asset. It has to be a list of length 2. Defaults to None.
    """

    input_location: Optional[list[float]] = Field(None, min_length=2, max_length=2)

    @model_validator(mode="before")
    @classmethod
    def check_locations(cls, data: Any) -> Any:
        assert "location" in data, "location has to be defined."
        if not "input_location" in data or not data["input_location"]:
            data["input_location"] = data["location"]
        return data

    def hash(self) -> str:
        """
        Returns a unique hash for the locatable asset considering its location and input location.

        Returns:
            str: Hash of the locatable asset.
        """
        return md5(
            (str(self.location) + str(self.input_location)).encode("utf-8")
        ).hexdigest()


class OutLocatable(Locatable):
    """
    Class that represents a locatable entity with an output location. The location parameter is the center point of the asset. It needs to be defined. The output location parameter specifies where products are removed from the asset. If not specified, the location is set as output location.

    Args:
        location (list[float]): The location of the locatable asset. It has to be a list of length 2.
        output_location (list[float], optional): The output location of the locatable asset. It has to be a list of length 2. Defaults to None.
    """

    output_location: Optional[list[float]] = Field(None, min_length=2, max_length=2)

    @model_validator(mode="before")
    @classmethod
    def check_locations(cls, data: Any) -> Any:
        assert "location" in data, "location has to be defined."
        if not "output_location" in data or not data["output_location"]:
            data["output_location"] = data["location"]
        return data

    def hash(self) -> str:
        """
        Returns a unique hash for the locatable asset considering its location and output location.

        Returns:
            str: Hash of the locatable asset.
        """
        return md5(
            (str(self.location) + str(self.output_location)).encode("utf-8")
        ).hexdigest()


class InOutLocatable(InLocatable, OutLocatable):
    """
    Class that represents a locatable entity with an input location and an output location. The location parameter is the center point of the asset. It needs to be defined. The input location parameter specifies where products are added to the asset. If not specified, the location is set as input location. The output location parameter specifies where products are removed from the asset. If not specified, the location is set as output location.

    Args:
        location (list[float]): The location of the locatable asset. It has to be a list of length 2.
        input_location (list[float], optional): The input location of the locatable asset. It has to be a list of length 2. Defaults to None.
        output_location (list[float], optional): The output location of the locatable asset. It has to be a list of length 2. Defaults to None.
    """

    @model_validator(mode="before")
    @classmethod
    def check_locations(cls, data: Any) -> Any:
        data = InLocatable.check_locations(data)
        data = OutLocatable.check_locations(data)
        return data

    def hash(self) -> str:
        """
        Returns a unique hash for the locatable asset considering its location, input location and output location.

        Returns:
            str: Hash of the locatable asset.
        """
        return md5(
            (
                str(self.location)
                + str(self.input_location)
                + str(self.output_location)
            ).encode("utf-8")
        ).hexdigest()



from abc import ABCMeta, abstractmethod
from copy import deepcopy
from collections import deque
from numpy import argmax

import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from prodsys.optimization.optimizer import Optimizer

from prodsys.optimization.optimization import evaluate
from prodsys.optimization.adapter_manipulation import mutation
from prodsys.optimization.optimization import check_valid_configuration
# from prodsys.optimization.util import document_individual

logger = logging.getLogger(__name__)


from pydantic import BaseModel, ConfigDict

from prodsys import adapters
from prodsys.optimization.util import (
    check_breakdown_states_available,
    create_default_breakdown_states,
)
from prodsys.util.util import set_seed


class TabuSearch:
    __metaclass__ = ABCMeta

    cur_steps = None

    tabu_size = None
    tabu_list = None

    initial_state = None
    current = None
    best = None

    max_steps = None
    max_score = None

    def __init__(self, initial_state, tabu_size, max_steps, max_score=None):
        self.initial_state = initial_state

        if isinstance(tabu_size, int) and tabu_size > 0:
            self.tabu_size = tabu_size
        else:
            raise TypeError("Tabu size must be a positive integer")

        if isinstance(max_steps, int) and max_steps > 0:
            self.max_steps = max_steps
        else:
            raise TypeError("Maximum steps must be a positive integer")

        if max_score is not None:
            if isinstance(max_score, (int, float)):
                self.max_score = float(max_score)
            else:
                raise TypeError("Maximum score must be a numeric type")

    def __str__(self):
        return (
            "TABU SEARCH: \n"
            + "CURRENT STEPS: %d \n"
            + "BEST SCORE: %f \n"
            + "BEST MEMBER: %s \n\n"
        ) % (self.cur_steps, self._score(self.best), str(self.best))

    def __repr__(self):
        return self.__str__()

    def _clear(self):
        """
        Resets the variables that are altered on a per-run basis of the algorithm

        :return: None
        """
        self.cur_steps = 0
        self.tabu_list = deque(maxlen=self.tabu_size)
        self.current = self.initial_state
        self.best = self.initial_state

    @abstractmethod
    def _score(self, state):
        pass

    @abstractmethod
    def _neighborhood(self):
        pass

    def _best(self, neighborhood):
        return neighborhood[argmax([self._score(x) for x in neighborhood])]

    def run(self, verbose=True):
        self._clear()
        for i in range(self.max_steps):
            self.cur_steps += 1

            neighborhood = self._neighborhood()
            neighborhood_best = self._best(neighborhood)

            while True:
                if all([x in self.tabu_list for x in neighborhood]):
                    return self.best, self._score(self.best)
                if neighborhood_best in self.tabu_list:
                    if self._score(neighborhood_best) > self._score(self.best):
                        self.tabu_list.append(neighborhood_best)
                        self.best = deepcopy(neighborhood_best)
                        break
                    else:
                        neighborhood.remove(neighborhood_best)
                        neighborhood_best = self._best(neighborhood)
                else:
                    self.tabu_list.append(neighborhood_best)
                    self.current = neighborhood_best
                    if self._score(self.current) > self._score(self.best):
                        self.best = deepcopy(self.current)
                    break

            if self.max_score is not None and self._score(self.best) > self.max_score:
                return self.best, self._score(self.best)
        return self.best, self._score(self.best)


class TabuSearchHyperparameters(BaseModel):
    """
    Hyperparameters for configuration optimization with tabu search.


    Args:
        seed (int): Seed for random number generator
        tabu_size (int): Size of tabu list
        max_steps (int): Maximum number of steps
        max_score (float): Maximum score
    """

    seed: int = 0
    tabu_size: int = 10
    max_steps: int = 300
    neighborhood_size: int = 10
    max_score: float = 500
    number_of_seeds: int = 1

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "seed": 0,
                    "tabu_size": 10,
                    "max_steps": 300,
                    "neighborhood_size": 10,
                    "max_score": 500,
                    "number_of_seeds": 1,
                },
            ]
        }
    )


def tabu_search_optimization(
    optimizer: "Optimizer",
):
    """
    Optimize a production system configuration using tabu search.

    Args:
        optimizer (Optimizer): The optimizer that contains the adapter, hyperparameters for tabu search, and initial solution (Defaults to None).
    """
    adapters.ProductionSystemAdapter.model_config["validate_assignment"] = False

    base_configuration = optimizer.adapter.model_copy(deep=True)
    if not adapters.check_for_clean_compound_processes(base_configuration):
        logger.warning(
            "Both compound processes and normal processes are used. This may lead to unexpected results."
        )
    if not check_breakdown_states_available(base_configuration):
        create_default_breakdown_states(base_configuration)

    hyper_parameters: TabuSearchHyperparameters = optimizer.hyperparameters
    set_seed(hyper_parameters.seed)

    performances = optimizer.performances_cache
    solution_dict = optimizer.optimization_cache_first_found_hashes

    class Algorithm(TabuSearch):

        def __init__(
            self,
            optimizer: "Optimizer",
            initial_state,
            tabu_size,
            max_steps,
            neighborhood_size,
            max_score=None,
        ):
            super().__init__(initial_state, tabu_size, max_steps, max_score)
            self.optimizer = optimizer
            self.previous_counter = None
            self.neighborhood_size = neighborhood_size

        def _score(self, state):
            fitness_values, event_log_dict = evaluate(
                base_scenario=base_configuration,
                solution_dict=solution_dict,
                number_of_seeds=hyper_parameters.number_of_seeds,
                adapter_object=state,
                full_save=optimizer.full_save
            )

            fitness_values, event_log_dict = self.optimizer.save_optimization_step(
                fitness_values=fitness_values,
                configuration=state,
                event_log_dict=event_log_dict,
            )
            performance = sum(
                [
                    value * weight
                    for value, weight in zip(fitness_values, self.optimizer.weights)
                ]
            )
            return performance

        def _neighborhood(self):
            neighboarhood = []
            for _ in range(self.neighborhood_size):
                while True:
                    configuration = mutation(individual=[deepcopy(self.current)])[0][0]
                    if check_valid_configuration(
                        configuration=configuration,
                        base_configuration=base_configuration,
                    ):
                        neighboarhood.append(configuration)
                        break
            return neighboarhood

    alg = Algorithm(
        optimizer=optimizer,
        initial_state=optimizer.initial_solutions,
        tabu_size=hyper_parameters.tabu_size,
        max_steps=hyper_parameters.max_steps,
        max_score=hyper_parameters.max_score,
        neighborhood_size=hyper_parameters.neighborhood_size
    )
    best_solution, best_objective_value = alg.run()
    optimizer.update_progress(
        num_steps=optimizer.progress.total_steps - optimizer.progress.completed_steps
    )



from typing import TYPE_CHECKING, Generator, List
from pydantic import BaseModel, ConfigDict, Field
from prodsys.adapters import adapter
from prodsys.factories import (
    process_factory,
    queue_factory,
    resource_factory,
    sink_factory,
)
from prodsys.models import queue_data, auxiliary_data

from prodsys.simulation import sim, router
from prodsys.simulation import auxiliary, logger


from prodsys.models import source_data


class AuxiliaryFactory(BaseModel):
    """
    Factory class for creating and managing auxiliary objects in the production system.
    """

    env: sim.Environment
    process_factory: process_factory.ProcessFactory
    queue_factory: queue_factory.QueueFactory
    resource_factory: resource_factory.ResourceFactory
    sink_factory: sink_factory.SinkFactory
    auxiliaries: List[auxiliary.Auxiliary] = []
    event_logger: logger.EventLogger = Field(default=False, init=False)
    auxiliary_counter: int = 0

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def create_auxiliary(self, adapter: adapter.ProductionSystemAdapter):
        """
        Create auxiliary objects based on the provided adapter's auxiliary data.

        Args:
            adapter (adapter.ProductionSystemAdapter): The adapter containing auxiliary data.
        """
        for auxiliary_data in adapter.auxiliary_data:
            for i, capacity_auxiliary_in_storage in enumerate(
                auxiliary_data.quantity_in_storages
            ):
                storage_id = auxiliary_data.storages[i]
                storage = next(
                    (
                        storage
                        for storage in adapter.queue_data
                        if storage.ID == storage_id
                    ),
                    None,
                )
                for _ in range(capacity_auxiliary_in_storage):
                    auxiliary = self.add_auxiliary(auxiliary_data, storage)
                    auxiliary.auxiliary_info.log_create_auxiliary(
                        resource=auxiliary.current_locatable,
                        _product=auxiliary,
                        event_time=self.env.now,
                    )

    def add_auxiliary(
        self,
        auxiliary_data: auxiliary_data.AuxiliaryData,
        storage: queue_data.StoreData,
    ) -> auxiliary.Auxiliary:
        """
        Add a new auxiliary object to the factory.

        Args:
                auxiliary_data (auxiliary_data.AuxiliaryData): The auxiliary data for the new object.
                storage (queue_data.QueueData): The storage data for the new object's current location.

        Returns:
                auxiliary.Auxiliary: The newly created auxiliary object.
        """
        values = {}
        auxiliary_data = auxiliary_data.model_copy(deep=True)
        auxiliary_data.ID = (
            str(auxiliary_data.auxiliary_type) + "_" + str(self.auxiliary_counter)
        )
        values.update({"env": self.env, "data": auxiliary_data})
        transport_process = self.process_factory.get_process(
            auxiliary_data.transport_process
        )
        values.update({"transport_process": transport_process})
        relevant_processes = []
        for process_id in auxiliary_data.relevant_processes:
            relevant_processes.append(self.process_factory.get_process(process_id))
        values.update({"relevant_processes": relevant_processes})
        relevant_transport_processes = []
        for transport_process_id in auxiliary_data.relevant_transport_processes:
            relevant_transport_processes.append(
                self.process_factory.get_process(transport_process_id)
            )
        values.update({"relevant_transport_processes": relevant_transport_processes})
        storage_from_queue_data = self.queue_factory.get_queue(storage.ID)
        values.update({"storage": storage_from_queue_data})
        router = self.get_router(
            source_data.RoutingHeuristic.FIFO
        )  # Add the routing_heuristic in auxiliary_data, like in source_data
        auxiliary_object = auxiliary.Auxiliary.model_validate(values)
        auxiliary_object.auxiliary_router = router
        auxiliary_object.current_locatable = storage_from_queue_data
        auxiliary_object.init_got_free()

        if self.event_logger:
            self.event_logger.observe_terminal_auxiliary_states(auxiliary_object)

        self.auxiliary_counter += 1
        self.auxiliaries.append(auxiliary_object)
        return auxiliary_object

    def place_auxiliaries_in_queues(self):
        """
        Place the auxiliary objects in the system.
        """
        for auxiliary in self.auxiliaries:
            auxiliary.current_locatable.put(auxiliary.product_data)

    def get_auxiliary(self, ID: str) -> auxiliary.Auxiliary:
        """
        Get the auxiliary object with the specified ID.

        Args:
            ID (str): The ID of the auxiliary object.

        Returns:
            auxiliary.Auxiliary: The auxiliary object with the specified ID.

        Raises:
            IndexError: If no auxiliary object with the specified ID is found.
        """
        return [s for s in self.auxiliaries if s.data.ID == ID].pop()

    def get_router(self, routing_heuristic: str):
        """
        Get the router object with the specified routing heuristic.

        Args:
            routing_heuristic (str): The routing heuristic to use.

        Returns:
            router.Router: The router object with the specified routing heuristic.
        """
        return router.Router(
            self.resource_factory,
            self.sink_factory,
            self,
            router.ROUTING_HEURISTIC[routing_heuristic],
        )

AuxiliaryFactory.model_rebuild()


from enum import Enum
from pydantic import BaseModel, ConfigDict
from typing import Literal, Union, Optional, List, Tuple, TYPE_CHECKING

from prodsys.models.performance_indicators import (
    KPIEnum,
    KPILevelEnum,
    KPI_UNION,
)


class Event(BaseModel):
    """
    Class that represents an event in the event log.

    Args:
        time (float): Time of the event.
        resource (str): Resource of the event.
        state (str): State of the event.
        state_type (str): State type of the event, representing a state.StateTypeEnum.
        activity (str): Activity of the event, representing a state.StateEnum.
        product (Optional[str], optional): Product of the event. Defaults to None.
        expected_end_time (Optional[float], optional): Expected end time of the event. Defaults to None.
        target_location (Optional[str], optional): Target location of the event. Defaults to None.
    """

    time: float
    resource: str
    state: str
    state_type: str
    activity: str
    product: Optional[str] = None
    expected_end_time: Optional[float] = None
    target_location: Optional[str] = None

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "time": 12.0,
                    "resource": "R1",
                    "state": "P1",
                    "state_type": "Production",
                    "activity": "start state",
                    "product": "Product_1_12",
                    "expected_end_time": 24.3,
                    "target_location": None,
                },
                {
                    "time": 24.3,
                    "resource": "R1",
                    "state": "P1",
                    "state_type": "Production",
                    "activity": "end state",
                    "product": "Product_1_12",
                    "expected_end_time": None,
                    "target_location": "L1",
                },
            ]
        }
    )


class Performance(BaseModel):
    """
    Class that represents the performance of a simulation run.

    Args:
        event_log (List[Event]): Event log of the simulation run.
        kpis (List[KPI_UNION]): List of KPIs of the simulation run.
    """

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "event_log": Event.model_config["json_schema_extra"]["examples"],
                    "kpis": [
                        {
                            "name": "throughput",
                            "target": "max",
                            "weight": 1,
                            "value": 4.32,
                            "context": ["system", "product_type"],
                            "product_type": "ProductType_1",
                        },
                        {
                            "name": "WIP",
                            "target": "min",
                            "weight": 1,
                            "value": 121,
                            "context": ["system", "product_type"],
                            "product_type": "ProductType_1",
                        },
                    ],
                }
            ]
        }
    )

    event_log: List[Event]
    kpis: List[KPI_UNION]

    def get_kpi_for_context(self, context: Tuple[KPILevelEnum, ...]) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if context == kpi.context]

    def get_kpi_for_name(self, name: KPIEnum) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if name == kpi.name]

    def get_kpi_for_context_and_name(
        self, context: Tuple[KPILevelEnum, ...], name: KPIEnum
    ) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if context == kpi.context and name == kpi.name]



from __future__ import annotations

import json
import time
from typing import TYPE_CHECKING
import warnings
import logging

from prodsys.optimization.optimization import evaluate_ea_wrapper
from prodsys.optimization.adapter_manipulation import (
    crossover,
    get_random_configuration_asserted,
    mutation,
    random_configuration_with_initial_solution,
)

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=RuntimeWarning)

from deap import algorithms, base, creator, tools
from pydantic import BaseModel, ConfigDict, Field

from prodsys.simulation import sim
from prodsys import adapters
from prodsys.optimization.util import (
    get_weights,
    check_breakdown_states_available,
    create_default_breakdown_states,
)
from prodsys.util.util import set_seed, run_from_ipython


if run_from_ipython():
    from multiprocessing.pool import ThreadPool as Pool
else:
    from multiprocessing.pool import Pool


if TYPE_CHECKING:
    from prodsys.optimization.optimizer import Optimizer


creator.create("FitnessMax", base.Fitness, weights=(1, 1, 1))  # als Tupel
creator.create("Individual", list, fitness=creator.FitnessMax)


class EvolutionaryAlgorithmHyperparameters(BaseModel):
    """
    Hyperparameters for configuration optimization using an evolutionary algorithm.

    Args:
        seed (int): Seed for the random number generator.
        number_of_generations (int): Number of generations to run the algorithm.
        population_size (int): Number of individuals in each generation.
        mutation_rate (float): Probability of mutating an individual.
        crossover_rate (float): Probability of crossover between two individuals.
        number_of_seeds (int): Number of seeds to use for simulation.
        number_of_processes (int): Number of processes to use for parallelization.
    """

    seed: int = Field(0, description="Seed for the random number generator.")
    number_of_generations: int = 10
    population_size: int = 10
    mutation_rate: float = 0.1
    crossover_rate: float = 0.1
    number_of_seeds: int = 1
    number_of_processes: int = 1

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "seed": 0,
                    "number_of_generations": 10,
                    "population_size": 10,
                    "mutation_rate": 0.1,
                    "crossover_rate": 0.1,
                    "number_of_seeds": 1,
                    "number_of_processes": 1,
                },
            ]
        }
    )


def register_functions_in_toolbox(
    base_configuration: adapters.JsonProductionSystemAdapter,
    solutions_dict: dict,
    weights: tuple,
    initial_solutions: list[adapters.JsonProductionSystemAdapter],
    hyper_parameters: EvolutionaryAlgorithmHyperparameters,
    full_save: bool,
):
    creator.create("FitnessMax", base.Fitness, weights=weights)  # als Tupel
    creator.create("Individual", list, fitness=creator.FitnessMax)
    toolbox = base.Toolbox()
    if initial_solutions:
        toolbox.register(
            "random_configuration",
            random_configuration_with_initial_solution,
            initial_solutions,
        )
    else:
        toolbox.register(
            "random_configuration",
            get_random_configuration_asserted,
            base_configuration,
        )
    toolbox.register(
        "individual",
        tools.initRepeat,
        creator.Individual,
        toolbox.random_configuration,
        n=1,
    )

    # Startpopulation erzeugen
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register(
        "evaluate",
        evaluate_ea_wrapper,
        base_configuration,
        solutions_dict,
        hyper_parameters.number_of_seeds,
        full_save
    )
    toolbox.register("mate", crossover)
    toolbox.register("mutate", mutation)

    toolbox.register("select", tools.selNSGA2)

    return toolbox


from prodsys.util import util


def evolutionary_algorithm_optimization(
    optimizer: "Optimizer",
):
    """
    Optimize a production system configuration using an evolutionary algorithm.

    Args:
        optimizer (Optimizer): The optimizer that contains the adapter, hyperparameters, and initial solutions.
    """
    adapters.ProductionSystemAdapter.model_config["validate_assignment"] = False

    base_configuration = optimizer.adapter.model_copy(deep=True)
    if not adapters.check_for_clean_compound_processes(base_configuration):
        logger.warning(
            "Both compound processes and normal processes are used. This may lead to unexpected results."
        )
    if not check_breakdown_states_available(base_configuration):
        create_default_breakdown_states(base_configuration)
    hyper_parameters: EvolutionaryAlgorithmHyperparameters = optimizer.hyperparameters

    set_seed(hyper_parameters.seed)

    start = time.perf_counter()

    toolbox = register_functions_in_toolbox(
        base_configuration=base_configuration,
        solutions_dict=optimizer.optimization_cache_first_found_hashes,
        weights=optimizer.weights,
        initial_solutions=optimizer.initial_solutions,
        hyper_parameters=hyper_parameters,
        full_save=optimizer.full_save,
    )

    population = toolbox.population(n=hyper_parameters.population_size)
    if hyper_parameters.number_of_processes > 1:
        pool = Pool(hyper_parameters.number_of_processes)
        toolbox.register("map", pool.map)
    else:
        toolbox.register("map", map)

    fitnesses = toolbox.map(toolbox.evaluate, population)
    for ind, (fit, event_log_dict) in zip(population, fitnesses):
        fit, event_log_dict = optimizer.save_optimization_step(
            fit, ind[0], event_log_dict
        )
        ind.fitness.values = fit
    population = toolbox.select(population, len(population))

    for g in range(hyper_parameters.number_of_generations):
        current_generation = g + 1
        optimizer.optimization_cache_first_found_hashes.current_generation = str(
            current_generation
        )
        # Vary population
        offspring = tools.selTournamentDCD(population, len(population))
        offspring = [toolbox.clone(ind) for ind in offspring]
        offspring = algorithms.varAnd(
            offspring,
            toolbox,
            cxpb=hyper_parameters.crossover_rate,
            mutpb=hyper_parameters.mutation_rate,
        )

        # Evaluate the individuals
        fitnesses = toolbox.map(toolbox.evaluate, offspring)
        for ind, fit_response in zip(offspring, fitnesses):
            fit, event_log_dict = fit_response
            fit, event_log_dict = optimizer.save_optimization_step(
                fit, ind[0], event_log_dict
            )
            ind.fitness.values = fit

        population = toolbox.select(
            population + offspring, hyper_parameters.population_size
        )
    if hyper_parameters.number_of_processes > 1:
        pool.close()



from typing import Union, Optional
from hashlib import md5

from pydantic import ConfigDict

from prodsys.models.core_asset import CoreAsset, InOutLocatable


class QueueData(CoreAsset):
    """
    Class that represents a queue for products at a resource / sink / source. If capacity is 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Args:
        ID (str): ID of the queue.
        description (str): Description of the queue.
        capacity (Union[int, float]): Capacity of the queue. If 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Examples:
        A finite queue with ID "Q1", description "Queue 1" and capacity 10:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
        )
        ```
        An infinite queue with ID "Q1", description "Queue 1" and capacity 0:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=0,
        )
        ```
    """

    capacity: Union[int, float] = 0.0

    def hash(self) -> str:
        """
        Returns a unique hash for the queue considering its capacity.


        Returns:
            str: Hash of the queue.
        """
        return md5((str(self.capacity)).encode("utf-8")).hexdigest()

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Q1",
                    "description": "Finte Queue",
                    "capacity": 10,
                },
                {
                    "ID": "Q1",
                    "description": "Infinite Queue",
                    "capacity": 0.0,
                },
            ]
        }
    )


class StoreData(QueueData, InOutLocatable):
    """
    Class that represents a store which is a queue with a loctation independent of a resource / source / sink.

    If capacity is 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Args:
        ID (str): ID of the queue.
        description (str): Description of the queue.
        capacity (Union[int, float]): Capacity of the queue. If 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.
        location (List[float]): Location of the queue. It has to be a list of length 2.
        input_location (Optional[List[float]], optional): Input location of the queue. Defaults to None.
        output_location (Optional[List[float]], optional): Output location of the queue. Defaults to None.

    Examples:
        A finite store with ID "Q1", description "Store Q1" and capacity 10:
        ``` py
        import prodsys
        prodsys.queue_data.StoreData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10, 0],
        )
        ```
        An infinite queue with ID "Q1", description "Queue 1" and capacity 0:
        ``` py
        import prodsys
        prodsys.queue_data.StoreData(
            ID="Q1",
            description="Queue 1",
            capacity=0,
            location=[10, 0],
            input_location=[11, 0],
            output_location=[12, 0],
        )
        ```
    """

    def hash(self) -> str:
        """
        Returns a unique hash for the queue considering its capacity.


        Returns:
            str: Hash of the queue.
        """
        queue_hash = QueueData.hash(self)
        location_hash = InOutLocatable.hash(self)
        return md5((location_hash + queue_hash).encode("utf-8")).hexdigest()

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Q1",
                    "description": "Finite Store",
                    "capacity": 10,
                    "location": [10, 0],
                },
                {
                    "ID": "Q1",
                    "description": "Infinite Store",
                    "capacity": 0.0,
                    "location": [10, 0],
                    "input_location": [11, 0],
                    "output_location": [12, 0],
                },
            ]
        }
    )


QUEUE_DATA_UNION = Union[QueueData, StoreData]



import json
import time
from copy import deepcopy
from typing import TYPE_CHECKING
from prodsys.util import util
from pydantic import BaseModel, ConfigDict
import logging

from prodsys.optimization.optimization import evaluate
from prodsys.optimization.adapter_manipulation import mutation
from prodsys.optimization.optimization import check_valid_configuration
# from prodsys.optimization.util import document_individual

logger = logging.getLogger(__name__)

from simanneal import Annealer

from prodsys.simulation import sim
from prodsys import adapters
from prodsys.optimization.util import (
    get_weights,
    check_breakdown_states_available,
    create_default_breakdown_states,
)
from prodsys.util.util import set_seed

if TYPE_CHECKING:
    from prodsys.optimization.optimizer import Optimizer


class ProductionSystemOptimization(Annealer):
    def __init__(
        self,
        optimizer: "Optimizer",
        base_configuration: adapters.ProductionSystemAdapter,
        performances: dict,
        solutions_dict: dict,
        start: float,
        weights: tuple,
        number_of_seeds: int = 1,
        initial_solution: adapters.ProductionSystemAdapter = None,
        full_save: bool = False,
    ):
        super().__init__(initial_solution, None)
        self.optimizer = optimizer
        self.save_folder = optimizer.save_folder
        self.base_configuration = base_configuration
        self.performances = performances
        self.solution_dict = solutions_dict
        self.start = start
        self.weights = weights
        self.number_of_seeds = number_of_seeds
        self.full_save = full_save
        self.previous_counter = None

    def default_update(self, step, T, E, acceptance, improvement):
        # ignore this function, it is only here to overwrite the print of the super class...
        pass

    def move(self):
        while True:
            configuration = mutation(individual=[deepcopy(self.state)])[0][0]
            if check_valid_configuration(
                configuration=configuration,
                base_configuration=self.base_configuration,
            ):
                self.state: adapters.ProductionSystemAdapter = configuration
                break

    def energy(self):
        fitness_values, event_log_dict = evaluate(
            base_scenario=self.base_configuration,
            solution_dict=self.solution_dict,
            number_of_seeds=self.number_of_seeds,
            adapter_object=self.state,
            full_save=self.full_save,
        )

        counter = len(self.performances["0"]) - 1
        if self.previous_counter is not None:
            fitness_values, event_log_dict = self.optimizer.save_optimization_step(
                fitness_values=fitness_values,
                configuration=self.state,
                event_log_dict=event_log_dict,
            )
        performance = sum(
            [value * weight for value, weight in zip(fitness_values, self.weights)]
        )
        self.previous_counter = counter
        return performance


class SimulatedAnnealingHyperparameters(BaseModel):
    """
    Hyperparameters to perform a configuration optimization with simulated annealing.
    Args:
        seed (int): Seed for random number generator
        Tmax (int): Maximum temperature
        Tmin (int): Minimum temperature
        steps (int): Number of steps
        updates (int): Number of updates
    """

    seed: int = 0
    Tmax: int = 10000
    Tmin: int = 1
    steps: int = 4000
    updates: int = 300
    number_of_seeds: int = 1

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "seed": 0,
                    "Tmax": 10000,
                    "Tmin": 1,
                    "steps": 4000,
                    "updates": 300,
                    "number_of_seeds": 1,
                },
            ]
        }
    )


def simulated_annealing_optimization(optimizer: "Optimizer"):
    """
    Optimize a production system configuration using simulated anealing.

    Args:
        base_configuration (adapters.ProductionSystemAdapter): production system to optimize.
        hyper_parameters (SimulatedAnnealingHyperparameters): Hyperparameters for simulated annealing.
        save_folder (str): Folder to save the results in. Defaults to "results".
        initial_solution (adapters.ProductionSystemAdapter, optional): Initial solution for optimization. Defaults to None.
    """
    adapters.ProductionSystemAdapter.model_config["validate_assignment"] = False
    base_configuration = optimizer.adapter.model_copy(deep=True)

    if not adapters.check_for_clean_compound_processes(base_configuration):
        logger.warning(
            "Both compound processes and normal processes are used. This may lead to unexpected results."
        )
    if not check_breakdown_states_available(base_configuration):
        create_default_breakdown_states(base_configuration)
    if not optimizer.initial_solutions:
        optimizer.initial_solutions = base_configuration.model_copy(deep=True)

    hyper_parameters: SimulatedAnnealingHyperparameters = optimizer.hyperparameters

    set_seed(hyper_parameters.seed)

    start = time.perf_counter()
    solutions_dict = optimizer.optimization_cache_first_found_hashes
    performances = optimizer.performances_cache

    pso = ProductionSystemOptimization(
        optimizer=optimizer,
        base_configuration=base_configuration,
        performances=performances,
        solutions_dict=solutions_dict,
        start=start,
        weights=optimizer.weights,
        number_of_seeds=hyper_parameters.number_of_seeds,
        initial_solution=optimizer.initial_solutions,
        full_save=optimizer.save_folder if optimizer.full_save else "",
    )

    pso.Tmax = hyper_parameters.Tmax
    pso.Tmin = hyper_parameters.Tmin
    pso.steps = hyper_parameters.steps
    pso.updates = hyper_parameters.updates

    internary, performance = pso.anneal()



from __future__ import annotations

from dataclasses import field
from typing import List, TYPE_CHECKING, Optional, List

from pydantic import ConfigDict, BaseModel, TypeAdapter

from prodsys.simulation import sim
from prodsys.factories import time_model_factory
from prodsys.models import state_data
from prodsys.simulation import state


if TYPE_CHECKING:
    from prodsys.adapters import adapter


class StateFactory:
    """
    Factory class that creates and stores `prodsys.simulation` state objects from `prodsys.models` state objects.

    Args:
        env (sim.Environment): prodsys simulation environment.
        time_model_factory (time_model_factory.TimeModelFactory): Factory that creates time model objects.
    """

    def __init__(
        self,
        env: sim.Environment,
        time_model_factory: time_model_factory.TimeModelFactory,
    ):
        self.env = env
        self.time_model_factory = time_model_factory
        self.state_data: List[state_data.STATE_DATA_UNION] = []
        self.states: List[state.STATE_UNION] = []

    def create_states_from_configuration_data(self, configuration_data: dict):
        for cls_name, items in configuration_data.items():
            for values in items.values():
                values.update({"type": cls_name})
                self.state_data.append(
                    TypeAdapter(state_data.STATE_DATA_UNION).validate_python(values)
                )
                self.add_state(self.state_data[-1])

    def get_loading_time_models_data(
        self, transport_state: state_data.STATE_DATA_UNION
    ) -> dict:
        if not isinstance(transport_state, state_data.TransportStateData):
            return {}
        loading_time_model_dict = {}
        if transport_state.loading_time_model_id is not None:
            loading_time_model_dict[
                "loading_time_model"
            ] = self.time_model_factory.get_time_model(
                transport_state.loading_time_model_id
            )
        if transport_state.unloading_time_model_id is not None:
            loading_time_model_dict[
                "unloading_time_model"
            ] = self.time_model_factory.get_time_model(
                transport_state.unloading_time_model_id
            )
        return loading_time_model_dict

    def get_repair_time_model_data(
        self, state_data: state_data.STATE_DATA_UNION
    ) -> dict:
        repair_time_model_dict = {}
        if (
            "repair_time_model_id" in state_data.model_dump()
            and state_data.model_dump()["repair_time_model_id"] is not None
        ):
            return {
                "repair_time_model": self.time_model_factory.get_time_model(
                    state_data.repair_time_model_id
                )
            }
        return repair_time_model_dict

    def get_battery_time_model_data(
        self, state_data: state_data.STATE_DATA_UNION
    ) -> dict:
        battery_time_model_dict = {}
        if (
            "battery_time_model_id" in state_data.model_dump()
            and state_data.model_dump()["battery_time_model_id"] is not None
        ):
            return {
                "battery_time_model": self.time_model_factory.get_time_model(
                    state_data.battery_time_model_id
                )
            }
        return battery_time_model_dict

    def add_state(self, state_data: state_data.STATE_DATA_UNION):
        values = {
            "state_data": state_data,
            "time_model": self.time_model_factory.get_time_model(
                state_data.time_model_id
            ),
            "env": self.env,
        }
        values.update(self.get_loading_time_models_data(state_data))
        values.update(self.get_repair_time_model_data(state_data))
        values.update(self.get_battery_time_model_data(state_data))
        self.states.append(TypeAdapter(state.STATE_UNION).validate_python(values))

    def create_states(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates state objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the state data.
        """
        for state_data in adapter.state_data:
            self.add_state(state_data)

    def get_states(self, IDs: List[str]) -> List[state.STATE_UNION]:
        """
        Returns a list of state objects with the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to sort the state objects.

        Returns:
            List[state.STATE_UNION]: List of state objects with the given IDs.
        """
        return [st for st in self.states if st.state_data.ID in IDs]



import logging

import importlib_metadata
import toml
from prodsys.conf.logging_config import set_logging
from prodsys.models import (
    processes_data,
    product_data,
    queue_data,
    auxiliary_data,
    resource_data,
    scenario_data,
    sink_data,
    source_data,
    state_data,
    time_model_data,
)
from prodsys import adapters
from prodsys.util import post_processing  # , optimization_util
from prodsys.simulation import runner

logger = logging.getLogger(__name__)


def get_version() -> str:
    try:
        return importlib_metadata.version("prodsys")
    except:
        logger.info(
            "Could not find version in package metadata. Trying to read from pyproject.toml"
        )
    try:
        pyproject = toml.load("pyproject.toml")
        return pyproject["tool"]["poetry"]["version"]
    except:
        logger.error(
            "Could not find pyproject.toml file. Trying to read from poetry.lock"
        )
    raise ModuleNotFoundError(
        "Could not find version in package metadata or pyproject.toml"
    )


VERSION = get_version()



from __future__ import annotations
from abc import ABC, abstractmethod
import json
import time
from typing import Any, Callable, Optional

import pandas as pd
from pydantic import TypeAdapter
from prodsys.adapters.json_adapter import JsonProductionSystemAdapter
from prodsys.optimization.optimization_data import (
    FitnessData,
    OptimizationProgress,
    OptimizationResults,
    OptimizationSolutions,
    SolutionMetadata,
    get_empty_optimization_results,
)
from prodsys.optimization.util import get_weights
from prodsys.util import util

if util.run_from_ipython():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from prodsys.adapters import ProductionSystemAdapter
from prodsys.optimization.evolutionary_algorithm import (
    EvolutionaryAlgorithmHyperparameters,
    evolutionary_algorithm_optimization,
)
from prodsys.optimization.math_opt import (
    MathOptHyperparameters,
    mathematical_optimization,
)
from prodsys.optimization.simulated_annealing import (
    SimulatedAnnealingHyperparameters,
    simulated_annealing_optimization,
)
from prodsys.optimization.tabu_search import (
    TabuSearchHyperparameters,
    tabu_search_optimization,
)

HyperParameters = (
    EvolutionaryAlgorithmHyperparameters
    | SimulatedAnnealingHyperparameters
    | TabuSearchHyperparameters
    | MathOptHyperparameters
)


class Optimizer(ABC):
    """
    Base Optimizer interface for optimizing the configuration of a production system.
    This class defines the common interface and shared logic for optimization but cannot be used directly.

    Note: Configurations are not stored within the FitnessData.
    Instead, they are cached separately by their hash in concrete subclasses.
    """

    def __init__(
        self,
        adapter: ProductionSystemAdapter,
        hyperparameters: HyperParameters,
        initial_solutions: Optional[list[ProductionSystemAdapter]] = None,
        full_save: bool = False,
    ) -> None:
        self.adapter = adapter
        self.hyperparameters = hyperparameters
        self.initial_solutions = initial_solutions
        self.full_save = full_save  # Determines whether event logs are saved

        # Do not cache configurations here; caching is implemented only in the concrete subclasses.
        self.weights = None
        self.pbar: Any = None

        self.optimization_cache_first_found_hashes = OptimizationSolutions()
        self.performances_cache: OptimizationResults = get_empty_optimization_results()
        self.progress = OptimizationProgress()
        self.start_time = None

    def get_algorithm_and_steps(self) -> tuple[Callable, int]:
        if isinstance(self.hyperparameters, EvolutionaryAlgorithmHyperparameters):
            updates = (
                self.hyperparameters.number_of_generations
                * self.hyperparameters.population_size
                + self.hyperparameters.population_size
            )
            self.weights = get_weights(self.adapter, "max")
            return evolutionary_algorithm_optimization, updates
        elif isinstance(self.hyperparameters, SimulatedAnnealingHyperparameters):
            updates = self.hyperparameters.steps
            self.weights = get_weights(self.adapter, "min")
            return simulated_annealing_optimization, updates
        elif isinstance(self.hyperparameters, TabuSearchHyperparameters):
            updates = self.hyperparameters.max_steps * 10  # 10 neighbors per step
            self.weights = get_weights(self.adapter, "max")
            return tabu_search_optimization, updates
        elif isinstance(self.hyperparameters, MathOptHyperparameters):
            self.weights = get_weights(self.adapter, "min")
            return mathematical_optimization, 1
        else:
            raise ValueError("No algorithm provided for the optimization.")

    def optimize(self):
        """
        Performs the optimization process.
        """
        if not self.hyperparameters:
            raise ValueError("No hyperparameters provided for the optimizer.")
        if not self.adapter:
            raise ValueError("No adapter provided for the optimizer.")
        self.progress = OptimizationProgress()
        self.optimization_cache_first_found_hashes = OptimizationSolutions()
        self.performances_cache = get_empty_optimization_results()
        self.start_time = time.perf_counter()

        algorithm, steps = self.get_algorithm_and_steps()
        self.progress.total_steps = steps
        self.pbar = tqdm(total=steps, desc="Optimization Progress", leave=True)
        self.pbar.update(0)
        algorithm(self)
        self.pbar.close()

    def save_optimization_step(
        self,
        fitness_values: Optional[list[float]],
        configuration: JsonProductionSystemAdapter,
        event_log_dict: Optional[dict] = None,
    ) -> tuple[list[float], dict]:
        """
        Save an optimization step, caching and persisting the fitness data.
        The configuration is stored separately and associated by its hash.
        """
        self.update_progress()
        fitness_data = self.get_fitness_data_entry(
            configuration, fitness_values, event_log_dict=event_log_dict
        )
        # Cache the configuration separately (only in subclasses that override cache_configuration)
        self.save_configuration(configuration)
        current_generation = (
            self.optimization_cache_first_found_hashes.current_generation
        )
        self.cache_fitness_data(fitness_data, current_generation, configuration)
        self.save_fitness_data(fitness_data, current_generation)
        return fitness_data.fitness, fitness_data.event_log_dict

    def get_fitness_data_entry(
        self,
        configuration: JsonProductionSystemAdapter,
        fitness_values: list[float],
        event_log_dict: Optional[dict],
    ) -> FitnessData:
        """
        Creates and returns a FitnessData entry for a given configuration.
        The configuration is not stored in the FitnessData but is cached separately.
        """
        objective_names = [
            obj.name.value for obj in configuration.scenario_data.objectives
        ]
        # If fitness_values and event_log_dict are both None, try to load from cache.
        if fitness_values is None and event_log_dict is None:
            adapter_object_hash = configuration.hash()
            if (
                adapter_object_hash
                not in self.optimization_cache_first_found_hashes.hashes
            ):
                raise ValueError(
                    f"Adapter hash {adapter_object_hash} not found in optimization cache for configuration {configuration.ID}. Error in saving."
                )
            evaluated_adapter_generation = (
                self.optimization_cache_first_found_hashes.hashes[
                    adapter_object_hash
                ].generation
            )
            evaluated_adapter_id = self.optimization_cache_first_found_hashes.hashes[
                adapter_object_hash
            ].ID
            fitness_data = self.performances_cache[evaluated_adapter_generation][
                evaluated_adapter_id
            ]
            event_log_dict = fitness_data.event_log_dict
            fitness_values = fitness_data.fitness

        agg_fitness = sum(
            value * weight for value, weight in zip(fitness_values, self.weights)
        )
        config_hash = configuration.hash()
        return FitnessData(
            agg_fitness=agg_fitness,
            fitness=fitness_values,
            time_stamp=time.perf_counter() - self.start_time,
            hash=config_hash,
            production_system=None,  # Do not store the configuration here
            objective_names=objective_names,
            event_log_dict=event_log_dict if self.full_save else None,
        )

    def cache_fitness_data(
        self,
        fitness_data: FitnessData,
        generation: str,
        configuration: JsonProductionSystemAdapter,
    ) -> None:
        """
        Caches the fitness data in memory.
        """
        if generation not in self.performances_cache:
            self.performances_cache[generation] = {}
        if fitness_data.hash not in self.optimization_cache_first_found_hashes.hashes:
            self.optimization_cache_first_found_hashes.hashes[fitness_data.hash] = (
                SolutionMetadata(generation=generation, ID=configuration.ID)
            )
        self.performances_cache[generation][configuration.ID] = fitness_data

    @abstractmethod
    def save_configuration(
        self, configuration: ProductionSystemAdapter
    ) -> None:
        """
        Caches the configuration. Must be implemented by concrete subclasses.
        """
        ...

    @abstractmethod
    def save_fitness_data(self, fitness_data: FitnessData, generation: str) -> None:
        """
        Persists the fitness data. Must be implemented by concrete subclasses.
        """
        ...

    def get_optimization_results(
        self, configuration_data: bool = False, event_log_data: bool = False
    ) -> OptimizationResults:
        """
        Retrieves the optimization results. If the cache is non-empty, it is returned.
        Otherwise, data is loaded from persistence.

        Flags:
         - configuration_data: load full configuration data if available. Default is True.
         - event_log_data: load the event log details if available. Default is False.
        """
        if self.performances_cache and not event_log_data:
            if not configuration_data and event_log_data == self.full_save:
                return self.performances_cache
            results = {}
            for generation, fitness_data_dict in self.performances_cache.items():
                results[generation] = {}
                for adapter_id, fitness_data in fitness_data_dict.items():
                    copied_fitness_data = fitness_data.model_copy(deep=True)
                    if not configuration_data:
                        copied_fitness_data.production_system = None
                    if not event_log_data:
                        copied_fitness_data.event_log_dict = None
                    copied_fitness_data.production_system = self.get_configuration_by_hash(
                        fitness_data.hash
                    )
                    results[generation][adapter_id] = copied_fitness_data
        else:
            results = self.get_optimization_results_from_persistence(
                configuration_data, event_log_data
            )
        return results

    def get_optimization_result_configuration(
        self, solution_id: str
    ) -> ProductionSystemAdapter:
        """
        Returns the configuration of the solution identified by solution_id.
        Retrieval is based on adapter id and generation.
        """
        optimization_results = self.get_optimization_results()
        for generation in optimization_results:
            for adapter_id, fitness_data in optimization_results[generation].items():
                if adapter_id == solution_id:
                    return self.get_configuration_by_hash(fitness_data.hash)
        raise ValueError(f"Solution {solution_id} not found in optimization results.")

    def get_configuration_by_adapter_id_and_generation(
        self, generation: str, adapter_id: str
    ) -> ProductionSystemAdapter:
        """
        Returns the configuration for a given generation and adapter id.
        """
        fitness_data = self.get_fitness_data(generation, adapter_id)
        return self.get_configuration_by_hash(fitness_data.hash)

    @abstractmethod
    def get_configuration_by_hash(
        self, configuration_hash: str
    ) -> ProductionSystemAdapter:
        """
        Retrieves the configuration based on its hash.
        Must be implemented by concrete subclasses.
        """
        ...

    @abstractmethod
    def get_optimization_results_from_persistence(
        self, configuration_data: bool = False, event_log_data: bool = False
    ) -> OptimizationResults:
        """
        Retrieves optimization results from persistence (e.g. file system).
        Must be implemented by concrete subclasses.
        """
        ...

    def get_fitness_data(self, generation: str, adapter_id: str) -> FitnessData:
        """
        Returns the fitness data for a given generation and adapter_id.
        """
        try:
            return self.get_fitness_data_from_cache(generation, adapter_id)
        except Exception:
            return self.get_fitness_data_from_persistence(generation, adapter_id)

    def get_fitness_data_from_cache(
        self, generation: str, adapter_id: str
    ) -> FitnessData:
        if (
            generation not in self.performances_cache
            or adapter_id not in self.performances_cache[generation]
        ):
            raise ValueError("Optimization result not found in cache.")
        return self.performances_cache[generation][adapter_id]

    @abstractmethod
    def get_fitness_data_from_persistence(
        self, generation: str, adapter_id: str
    ) -> FitnessData:
        """
        Retrieves fitness data from persistence. Must be implemented by concrete subclasses.
        """
        ...

    def update_progress(self, num_steps: int = 1) -> None:
        """
        Updates the progress bar.
        """
        if self.pbar:
            self.pbar.update(num_steps)
            self.progress.completed_steps += num_steps
        else:
            raise ValueError("Progress bar not initialized.")

    def get_progress(self) -> OptimizationProgress:
        return self.progress

    @abstractmethod
    def save_configuration(
        self, configuration_hash: str, configuration: ProductionSystemAdapter
    ) -> None:
        """
        Caches the configuration.
        The base class does not implement caching.
        Subclasses with caching enabled should override this method.
        """
        ...


class InMemoryOptimizer(Optimizer):
    """
    Optimizer implementation that holds all data in memory (cache).
    No persistence is performed.
    """

    def __init__(
        self,
        adapter: ProductionSystemAdapter,
        hyperparameters: HyperParameters,
        initial_solutions: Optional[list[ProductionSystemAdapter]] = None,
        full_save: bool = False,
    ) -> None:
        super().__init__(adapter, hyperparameters, initial_solutions, full_save)
        self.configuration_cache: dict[str, ProductionSystemAdapter] = {}

    def save_configuration(
        self, configuration: ProductionSystemAdapter
    ) -> None:
        configuration_hash = configuration.hash()
        self.configuration_cache[configuration_hash] = configuration

    def save_fitness_data(self, fitness_data: FitnessData, generation: str) -> None:
        # Here also add the configuration to the cached fitness data
        configuration = self.configuration_cache.get(fitness_data.hash)
        if configuration is None:
            raise ValueError(
                f"Configuration with hash {fitness_data.hash} not found in cache."
            )
        fitness_data.production_system = configuration

    def get_configuration_by_hash(
        self, configuration_hash: str
    ) -> ProductionSystemAdapter:
        if configuration_hash not in self.configuration_cache:
            raise ValueError("Configuration not found in cache.")
        return self.configuration_cache[configuration_hash]

    def get_optimization_results_from_persistence(
        self, configuration_data: bool = False, event_log_data: bool = False
    ) -> OptimizationResults:
        # Since all data is in memory, simply return the cache.
        return self.performances_cache

    def get_fitness_data_from_persistence(
        self, generation: str, adapter_id: str
    ) -> FitnessData:
        raise NotImplementedError(
            "Method get_fitness_data_from_persistence not implemented in InMemoryOptimizer. "
            "Ensure the optimization run was conducted and use get_fitness_data instead."
        )


class FileSystemSaveOptimizer(Optimizer):
    """
    Optimizer implementation that saves data to the file system.

    Args:
        adapter (ProductionSystemAdapter): The production system adapter to optimize.
        hyperparameters (HyperParameters): The hyperparameters for the optimization.
        save_folder (str): The folder where data will be saved.
        initial_solutions (Optional[list[ProductionSystemAdapter]], optional): Initial solutions to start the optimization. Defaults to None.
        full_save (bool, optional): Whether to save full event log data. Defaults to False.
    """

    def __init__(
        self,
        adapter: ProductionSystemAdapter,
        hyperparameters: HyperParameters,
        save_folder: str,
        initial_solutions: Optional[list[ProductionSystemAdapter]] = None,
        full_save: bool = False,
    ) -> None:
        super().__init__(adapter, hyperparameters, initial_solutions, full_save)
        self.save_folder = save_folder
        self.configuration_cache: dict[str, ProductionSystemAdapter] = {}
        util.prepare_save_folder(self.save_folder + "/")

    def save_configuration(
        self, configuration: ProductionSystemAdapter
    ) -> None:
        configuration_hash = configuration.hash()
        JsonProductionSystemAdapter.model_validate(configuration).write_data(
            f"{self.save_folder}/hash_{configuration_hash}.json"
        )

    def save_fitness_data(self, fitness_data: FitnessData, generation: str) -> None:
        """
        Saves the fitness data to disk.
        If full_save is True, event log data is also persisted.
        """
        if self.full_save and fitness_data.event_log_dict:
            df = pd.DataFrame.from_dict(fitness_data.event_log_dict)
            df.to_json(
                f"{self.save_folder}/event_log_dict_{fitness_data.hash}.json", indent=4
            )
            fitness_data.event_log_dict = None  # Free up memory
        # Update the aggregated optimization results on disk.
        optimization_results = {}
        for gen in self.performances_cache:
            if gen not in optimization_results:
                optimization_results[gen] = {}
            for adapter_id, fitness_data_entry in self.performances_cache[gen].items():
                optimization_results[gen][adapter_id] = fitness_data_entry.model_dump(
                    exclude={"production_system", "event_log_dict"}
                )
        with open(f"{self.save_folder}/optimization_results.json", "w") as json_file:
            json.dump(optimization_results, json_file, indent=4)

    def get_configuration_by_hash(
        self, configuration_hash: str
    ) -> ProductionSystemAdapter:
        config = JsonProductionSystemAdapter()
        config.read_data(
            f"{self.save_folder}/hash_{configuration_hash}.json"
        )
        return config

    def get_fitness_data_from_persistence(
        self, generation: str, adapter_id: str
    ) -> FitnessData:
        """
        Loads the fitness data for a given generation and adapter_id from disk.
        """
        with open(f"{self.save_folder}/optimization_results.json", "r") as json_file:
            optimization_results = json.load(json_file)
        fitness_data_dict = optimization_results[generation][adapter_id]
        fitness_data = FitnessData.model_validate(fitness_data_dict)
        # In this design, the configuration is loaded separately.
        fitness_data.production_system = None
        if self.full_save:
            df = pd.read_json(
                f"{self.save_folder}/event_log_dict_{fitness_data.hash}.json"
            )
            fitness_data.event_log_dict = df.to_dict()
        else:
            fitness_data.event_log_dict = None
        return fitness_data

    def get_optimization_results_from_persistence(
        self, configuration_data: bool = False, event_log_data: bool = False
    ) -> OptimizationResults:
        """
        Loads the entire optimization results from disk.
        The flags determine if full configuration data and/or event log data are included.
        """
        with open(f"{self.save_folder}/optimization_results.json", "r") as json_file:
            optimization_results_data = json.load(json_file)
        optimization_results: OptimizationResults = TypeAdapter(
            OptimizationResults
        ).validate_python(optimization_results_data)
        if configuration_data:
            for generation in optimization_results:
                for adapter_id, fitness_data in optimization_results[
                    generation
                ].items():
                    config = self.get_configuration_by_hash(fitness_data.hash)
                    config.ID = adapter_id
                    self.configuration_cache[fitness_data.hash] = config
                    fitness_data.production_system = config
        if event_log_data and self.full_save:
            for generation in optimization_results:
                for adapter_id, fitness_data in optimization_results[
                    generation
                ].items():
                    df = pd.read_json(
                        f"{self.save_folder}/event_log_dict_{fitness_data.hash}.json"
                    )
                    fitness_data.event_log_dict = df.to_dict()
        return optimization_results



from __future__ import annotations

from typing import Literal

import pandas as pd

import numpy as np

import logging

logger = logging.getLogger(__name__)

WARM_UP_CUT_OFF = 0.15


def mser5(data: pd.DataFrame, column: str) -> int:
    """
    Implements the MSER-5 method to determine the optimal truncation point
    for a given column in a DataFrame.

    Args:
        data (pd.DataFrame): The input DataFrame containing simulation output data.
        column (str): The name of the column to analyze.

    Returns:
        int: The index of the optimal truncation point.
    """
    if column not in data.columns:
        raise ValueError(f"Column '{column}' does not exist in the DataFrame.")

    if data[column].isnull().any():
        raise ValueError("Input column contains NaN values. Please clean the data.")

    values = data[column].to_numpy()
    n = len(values)

    if n < 5:
        return n
    # FIXME: this method is not working properly...
    # Divide data into batches of size 5
    batched_data = [
        values[i : i + 5] for i in range(0, n, 5) if len(values[i : i + 5]) == 5
    ]
    batch_means = np.array([np.mean(batch) for batch in batched_data])

    # Find truncation point that minimizes the standard error
    min_std_error = float("inf")
    optimal_index = None

    for i in range(len(batch_means)):
        truncated_means = batch_means[i:]
        std_error = np.std(truncated_means) / np.sqrt(len(truncated_means))

        if std_error < min_std_error:
            min_std_error = std_error
            optimal_index = i * 5  # Convert batch index back to original data index
    if optimal_index is None or optimal_index == (len(batch_means) - 1) * 5:
        return n
    return optimal_index


def threshold_stabilization(df: pd.DataFrame, column: str) -> int:
    """
    Determines the warm-up cutoff for steady-state simulations using the Method of Winter.

    Parameters:
        df (pd.DataFrame): The DataFrame containing the time series data.
        column (str): The name of the column with the performance metric data.

    Returns:
        int: Index of the warm-up cutoff.
    """
    # Extract the data for the specified column
    data = df[column].values

    # Compute the cumulative averages
    cumulative_avg = np.cumsum(data) / np.arange(1, len(data) + 1)

    # Compute moving range to detect stabilization
    moving_range = np.abs(np.diff(cumulative_avg))

    # Threshold for stabilization (e.g., small variation indicates steady state)
    threshold = np.mean(moving_range) / 10  # Adjust threshold as needed

    # Find the cutoff index where stabilization starts
    for i in range(1, len(moving_range)):
        if all(moving_range[i:] < threshold):
            cutoff_index = i
            break
    else:
        cutoff_index = len(data)  # If no steady state is found, use full data

    return cutoff_index


def static_ratio(df: pd.DataFrame, column: str) -> int:
    """
    Determines the warm-up cutoff for steady-state simulations using a static cutoff.

    Parameters:
        df (pd.DataFrame): The DataFrame containing the time series data.
        column (str): The name of the column with the performance metric data.

    Returns:
        int: Index of the warm-up cutoff.
    """
    data = df[column].values
    cutoff_index = int(len(data) * WARM_UP_CUT_OFF)
    return cutoff_index


def get_warm_up_cutoff_index(
    df: pd.DataFrame,
    column: str,
    method: Literal["mser5", "threshold_stabilization", "static_ratio"] = "mser5",
) -> int:
    """
    Calculates the warm up cutoff time for the simulation results.

    Args:
        df (pd.DataFrame): Data frame with the simulation results.
        max_time_column_name (str): The column where the maximum value is used to determine the warm up cutoff time.

    Returns:
        float: Warm up cutoff indexfor the input data frame and column.
    """
    if method == "mser5":
        cut_off_index = mser5(df, column)
    elif method == "threshold_stabilization":
        cut_off_index = threshold_stabilization(df, column)
    elif method == "static_ratio":
        cut_off_index = static_ratio(df, column)
    else:
        raise ValueError(
            "Invalid method specified for warm up cutoff index calculation. Please choose from 'mser5', 'threshold_stabilization', or 'static_ratio'."
        )

    return cut_off_index



import os
from typing import List
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots

from prodsys.util import post_processing


def hex_to_rgba(h, alpha):
    return tuple([int(h.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4)] + [alpha])


def plot_throughput_time_distribution(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the throughput time distribution of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df_tp = post_processor.df_throughput
    grouped = df_tp.groupby(by="Product_type")["Throughput_time"].apply(list)

    values = grouped.values

    group_labels = grouped.index

    # Create distplot with custom bin_size
    fig = ff.create_distplot(
        values, group_labels, bin_size=0.2, show_curve=True, show_hist=False
    )
    fig.update_layout(
        xaxis_title="Throughput Time [Minutes]",
        yaxis_title="Probability Density",
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "throughput_time_distribution.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, include_plotlyjs="cdn", full_html=False)
    if return_image:
        image_path = os.path.join(
            os.getcwd(), "plots", "throughput_time_distribution.png"
        )
        fig.write_image(image_path)
        return image_path


def plot_line_balance_kpis(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots line balancing key performance indicators (throughput, WIP, and Throughput Time) after total time.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """

    df_output = post_processor.get_aggregated_data()
    fig = make_subplots(
        rows=1,
        cols=3,
        specs=[[{"type": "indicator"}, {"type": "indicator"}, {"type": "indicator"}]],
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=round(np.sum(list(df_output["Throughput"]["Output"].values()))),
            title={"text": "Total Output [Products]"},
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=round(np.mean(list(df_output["WIP"]["WIP"].values()))),
            title={"text": "Average Work In Progress (WIP) [Products]"},
        ),
        row=1,
        col=2,
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=round(
                np.nanmean(
                    list(df_output["Throughput time"]["Throughput_time"].values())
                )
            ),
            title={"text": "Average Throughput Time [Minutes]"},
        ),
        row=1,
        col=3,
    )

    fig.update_layout(
        title_text="Line Balancing KPIs after Total Time",
        annotations=[
            dict(
                x=0.5,
                y=-0.1,
                showarrow=False,
                text="All KPI's are calculated for the steady state production",
                xref="paper",
                yref="paper",
            )
        ],
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "line_balance.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=False)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "line_balance_kpis.png")
        fig.write_image(image_path)
        return image_path


def plot_oee(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the Overall Equipment Effectiveness (OEE) with Availability, Performance & Quality using the given post_processor.

    Args:
        post_processor: An instance of the post_processing.PostProcessor class.
    """
    df_oee = post_processor.df_oee_production_system
    fig = make_subplots(
        rows=1,
        cols=4,
        specs=[
            [
                {"type": "indicator"},
                {"type": "indicator"},
                {"type": "indicator"},
                {"type": "indicator"},
            ]
        ],
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=df_oee["Value"][0],
            title={"text": "Availability in %"},
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=df_oee["Value"][1],
            title={"text": "Performance in %"},
        ),
        row=1,
        col=2,
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=df_oee["Value"][2],
            title={"text": "Quality in %"},
        ),
        row=1,
        col=3,
    )

    fig.add_trace(
        go.Indicator(
            mode="number",
            value=df_oee["Value"][3],
            title={"text": "OEE in %"},
        ),
        row=1,
        col=4,
    )

    fig.update_layout(
        title_text="Overall Equipment Effectiveness (OEE)",
        annotations=[
            dict(
                x=0.5,
                y=-0.1,
                showarrow=False,
                text="OEE = Availability * Performance * Quality",
                xref="paper",
                yref="paper",
            )
        ],
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "oee.html"), auto_open=not return_html
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "oee.png")
        fig.write_image(image_path)
        return image_path


def plot_production_flow_rate_per_product(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the production flow rate per product.

    Args:
        post_processor (post_processing.PostProcessor): The post processor object containing the data.
    """
    percentage_df = post_processor.df_production_flow_ratio

    fig = go.Figure(
        data=[
            go.Bar(
                name="Production",
                y=percentage_df["Product_type"],
                x=percentage_df["Production "],
                marker_color="steelblue",
                orientation="h",
                text=percentage_df["Production "].round(2),
                textposition="auto",
                textangle=-90,
            ),
            go.Bar(
                name="Transport",
                y=percentage_df["Product_type"],
                x=percentage_df["Transport "],
                marker_color="darkseagreen",
                orientation="h",
                text=percentage_df["Transport "].round(2),
                textposition="auto",
                textangle=-90,
            ),
            go.Bar(
                name="Idle",
                y=percentage_df["Product_type"],
                x=percentage_df["Idle "],
                marker_color="lightcoral",
                orientation="h",
                text=percentage_df["Idle "].round(2),
                textposition="auto",
                textangle=-90,
            ),
        ]
    )
    # Change the bar mode
    fig.update_layout(barmode="stack")
    fig.update_xaxes(title_text="Percentage [%]")
    fig.update_layout(
        title_text="Production Flow Rate (PFO) per Product",
        annotations=[
            dict(
                x=0.5,
                y=-0.1,
                showarrow=False,
                text="Time spent of product in Production, Transport and Idle during the whole throughput time",
                xref="paper",
                yref="paper",
            )
        ],
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "production_flow_rate_product_type.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(
            os.getcwd(), "plots", "production_flow_rate_per_product.png"
        )
        fig.write_image(image_path)
        return image_path


def plot_boxplot_resource_utilization(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots a boxplot to visualize resource utilization per station.

    Args:
        post_processor (post_processing.PostProcessor): The post processor object containing the data.
    """
    df_time_per_state = post_processor.df_aggregated_resource_bucket_states_boxplot
    resources = df_time_per_state["Resource"].unique()
    df_productive_time = df_time_per_state.loc[df_time_per_state["Time_type"] == "PR"]
    fig = go.Figure()

    for resource in resources:
        df_resource = df_productive_time.loc[df_time_per_state["Resource"] == resource]

        if len(df_resource) == 0:
            df_resource = pd.DataFrame({"Resource": [resource], "percentage": [0]})

        fig.add_trace(
            go.Box(
                y=df_resource["percentage"],
                name=f"{resource}",
                boxmean=True,
            )
        )

    fig.update_layout(
        title_text="Utilization per Station",
        yaxis_title="Percentage [%]",
        showlegend=False,
        annotations=[
            dict(
                x=0.5,
                y=-0.7,
                showarrow=False,
                # text="Dashed Line = Mean & Solid Line = Median, Whiskers = Q1/Q3 +/- 1.5 * IQR(Q3-Q1)",
                xref="paper",
                yref="paper",
            )
        ],
        height=600,
    )
    fig.update_yaxes(range=[0, 100])

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "resource_box_plots.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(
            os.getcwd(), "plots", "boxplot_resource_utilization.png"
        )
        fig.write_image(image_path)
        return image_path


def plot_throughput_time_over_time(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the throughput time over time of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df_tp = post_processor.df_throughput

    fig = px.scatter(
        df_tp,
        x="Start_time",
        y="Throughput_time",
        color="Product_type",
        trendline="expanding",
    )
    fig.data = [t for t in fig.data if t.mode == "lines"]
    fig.update_traces(showlegend=True)
    fig.update_layout(
        xaxis_title="Start Time [Minutes]",
        yaxis_title="Throughput Time [Minutes]",
    )
    min_start_time = df_tp["Start_time"].min()
    max_start_time = df_tp["Start_time"].max()

    warum_up_cut_off_time = post_processor.warm_up_cutoff_time

    new_x_range = [min_start_time, max_start_time]

    fig.update_layout(xaxis_range=new_x_range)

    fig.add_vline(
        x=warum_up_cut_off_time,
        line_dash="dash",
        line_color="red",
        annotation_text="Steady State",
        annotation_position="top right",
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "throughput.html"), auto_open=not return_html
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "throughput_time_over_time.png")
        fig.write_image(image_path)
        return image_path


def plot_time_per_state_of_resources(
    post_processor: post_processing.PostProcessor,
    normalized: bool = True,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the time per state of the resources of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
        normalized (bool, optional): If True, the time per state is normalized with the total time of the simulation. Defaults to True.
    """
    df_time_per_state = post_processor.df_aggregated_resource_states

    if normalized:
        y_column = "percentage"
    else:
        y_column = "time_increment"

    fig = px.bar(
        df_time_per_state,
        x="Resource",
        y=y_column,
        color="Time_type",
        color_discrete_map={
            "PR": "green",
            "SB": "yellow",
            "UD": "red",
            "ST": "blue",
            "CR": "grey",
        },
    )
    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "resource_states.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(
            os.getcwd(), "plots", "time_per_state_of_resources.png"
        )
        fig.write_image(image_path)
        return image_path


def plot_util_WIP_resource(
    post_processor: post_processing.PostProcessor,
    normalized: bool = True,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the time per state of the resources of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
        normalized (bool, optional): If True, the time per state is normalized with the total time of the simulation. Defaults to True.
    """
    df_time_per_state = post_processor.df_mean_wip_per_station
    df_time_per_state["mean_wip"] = np.maximum(
        np.ceil(df_time_per_state["mean_wip"]), 1
    )
    fig1 = go.Figure()
    fig1.add_trace(
        go.Bar(
            name="mean_wip",
            x=df_time_per_state["Resource"],
            y=df_time_per_state["mean_wip"],
            marker_color="purple",
            yaxis="y2",
        )
    )

    df_time_per_state2 = post_processor.df_aggregated_resource_bucket_states
    df_time_per_state2 = df_time_per_state2.loc[df_time_per_state2["Time_type"] == "PR"]

    resources = df_time_per_state2["Resource"].unique()
    fig2 = go.Figure()
    for resource in resources:
        df_resource = df_time_per_state2.loc[df_time_per_state2["Resource"] == resource]
        fig2.add_trace(
            go.Box(
                y=df_resource["percentage"],
                name=f"{resource}",
                boxmean=True,  # mean and standard deviation
            )
        )

    fig2.update_xaxes(categoryorder="array", categoryarray=resources)
    fig1.update_xaxes(categoryorder="array", categoryarray=resources)

    fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.1)
    for trace in fig1.data:
        fig.add_trace(trace, row=2, col=1)
    for trace in fig2.data:
        fig.add_trace(trace, row=1, col=1)

    fig.update_layout(
        title="Mean WIP and Utilization per Station",
        showlegend=False,
        height=800,  # adjust height if needed
    )

    fig.update_yaxes(title_text="Average WIP [Products]", row=2, col=1)
    fig.update_yaxes(title_text="Percentage [%]", row=1, col=1)

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "mean_wip_util_station.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "util_WIP_resource.png")
        fig.write_image(image_path)
        return image_path


def plot_transport_utilization_over_time(
    post_processor: post_processing.PostProcessor,
    transport_resource_names: List[str],
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the utilization of the transport_agv resource over time.

    Args:
        post_processor (post_processing.PostProcessor): The post processor object containing the data.
        transport_resource_names (List[str]): List of names of the transport resources.
    """
    df_time_per_state = post_processor.df_aggregated_resource_bucket_states
    transport_resource_names = set(transport_resource_names)
    df_agv_pr = df_time_per_state.loc[
        (df_time_per_state["Time_type"] == "PR")
        & (df_time_per_state["Resource"].isin(transport_resource_names))
    ]
    fig = go.Figure()
    for resource in transport_resource_names:
        df_agv_pr_resource = df_agv_pr.loc[df_agv_pr["Resource"] == resource]
        fig.add_trace(
            go.Scatter(
                x=df_agv_pr_resource["Time"],
                y=df_agv_pr_resource["percentage"],
                mode="lines",
                name=resource,
                #    line=dict(shape='spline', smoothing=2),  # Apply smoothing
                #    line=dict(shape='hv'),  # Apply smoothing
            ),
        )

    fig.update_layout(
        title="AGV Utilization Over Time",
        xaxis_title="Time in Minutes",
        yaxis_title="Percentage [%]",
    )

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "Transport_utilization_over_time.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(
            os.getcwd(), "plots", "transport_utilization_over_time.png"
        )
        fig.write_image(image_path)
        return image_path


def plot_WIP_with_range(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the WIP of the production system over time of the simulation with a range of the WIP based on a standard deviation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df = post_processor.df_WIP.copy()
    fig = px.scatter(df, x="Time", y="WIP")
    df["Product_type"] = "Total"

    df_per_product = post_processor.df_WIP_per_product.copy()

    df = pd.concat([df, df_per_product])

    fig = go.Figure()

    window = 5000

    def get_colors() -> List[str]:
        return (
            px.colors.qualitative.Dark24
            + px.colors.qualitative.Light24
            + px.colors.qualitative.G10
        )

    colors = get_colors()

    for product_type, df_product_type in df.groupby(by="Product_type"):
        df_product_type["WIP_avg"] = (
            df_product_type["WIP"].rolling(window=window).mean()
        )
        df_product_type["WIP_std"] = df_product_type["WIP"].rolling(window=window).std()
        if not colors:
            colors = get_colors()
        color = colors.pop(0)
        fig.add_scatter(
            name=product_type,
            x=df_product_type["Time"],
            y=df_product_type["WIP_avg"],
            mode="lines",
            line=dict(color=color),
        )
        fig.add_scatter(
            name=product_type + " Upper Bound",
            x=df_product_type["Time"],
            y=df_product_type["WIP_avg"] + df_product_type["WIP_std"],
            mode="lines",
            line=dict(dash="dash", color=color),
            showlegend=False,
        )
        fig.add_scatter(
            name=product_type + " Lower Bound",
            x=df_product_type["Time"],
            y=df_product_type["WIP_avg"] - df_product_type["WIP_std"],
            mode="lines",
            line=dict(dash="dash", color=color),
            fill="tonexty",
            fillcolor="rgba" + str(hex_to_rgba(color, 0.2)),
            showlegend=False,
        )
    fig.update_layout(
        xaxis_title="Time [Minutes]",
        yaxis_title="WIP [Products]",
    )

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "WIP_with_range.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "WIP_with_range.png")
        fig.write_image(image_path)
        return image_path


def plot_WIP(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the WIP of the production system over time of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df = post_processor.df_WIP.copy()
    fig = px.scatter(df, x="Time", y="WIP")
    df["Product_type"] = "Total"

    df_per_product = post_processor.df_WIP_per_product.copy()

    df = pd.concat([df, df_per_product])
    fig = px.scatter(
        df,
        x="Time",
        y="WIP",
        color="Product_type",
        trendline="expanding",
        opacity=0.01,
    )
    fig.data = [t for t in fig.data if t.mode == "lines"]
    fig.update_traces(showlegend=True)
    fig.update_layout(
        xaxis_title="Time [Minutes]",
        yaxis_title="WIP [Products]",
    )

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "WIP.html"), auto_open=not return_html
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "WIP.png")
        fig.write_image(image_path)
        return image_path


def plot_auxiliary_WIP(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the WIP of the production system over time of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df = post_processor.df_auxiliary_WIP.copy()
    fig = px.scatter(df, x="Time", y="auxiliary_WIP")
    df["Auxiliary_type"] = "Total"

    df_per_product = post_processor.df_auxiliary_WIP_per_auxiliary_type.copy()

    df = pd.concat([df, df_per_product])
    fig = px.scatter(
        df,
        x="Time",
        y="auxiliary_WIP",
        color="Auxiliary_type",
        trendline="expanding",
        opacity=0.01,
    )
    fig.data = [t for t in fig.data if t.mode == "lines"]
    fig.update_traces(showlegend=True)
    fig.update_layout(
        xaxis_title="Time [Minutes]",
        yaxis_title="Auxiliary WIP [Auxiliaries]",
    )

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "WIP.html"), auto_open=not return_html
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "WIP.png")
        fig.write_image(image_path)
        return image_path


def plot_WIP_per_resource(
    post_processor: post_processing.PostProcessor,
    return_html: bool = False,
    return_image: bool = False,
):
    """
    Plots the WIP of the production system and the resources in the production system over time of the simulation.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    df = post_processor.df_WIP.copy()
    fig = px.scatter(df, x="Time", y="WIP")
    df["Resource"] = "Total"

    df_per_resource = post_processor.df_WIP_per_resource.copy()
    df_per_resource["Resource"] = df_per_resource["WIP_resource"]

    df = pd.concat([df, df_per_resource])
    fig = px.scatter(
        df,
        x="Time",
        y="WIP",
        color="Resource",
        trendline="expanding",
        opacity=0.01,
    )
    fig.data = [t for t in fig.data if t.mode == "lines"]
    fig.update_traces(showlegend=True)
    fig.update_layout(
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=-0.3,
            xanchor="center",
            x=0.5,
        ),
        xaxis_title="Resource",
        yaxis_title="WIP [Products]",
    )

    if not os.path.exists(os.path.join(os.getcwd(), "plots")):
        os.makedirs(os.path.join(os.getcwd(), "plots"))
    fig.write_html(
        os.path.join(os.getcwd(), "plots", "WIP_per_resource.html"),
        auto_open=not return_html,
    )

    if return_html:
        return pio.to_html(fig, full_html=True)
    if return_image:
        image_path = os.path.join(os.getcwd(), "plots", "WIP_per_resource.png")
        fig.write_image(image_path)
        return image_path


def print_aggregated_data(post_processor: post_processing.PostProcessor):
    """
    Prints the aggregated data of the simulation, comprising the throughput, WIP, throughput time and resource states.

    Args:
        post_processor (post_processing.PostProcessor): Post processor of the simulation.
    """
    print("\n------------- Throughput -------------\n")

    print(post_processor.df_aggregated_output_and_throughput)

    print("------------- WIP -------------\n")
    print(post_processor.df_aggregated_WIP)

    if post_processor.get_auxiliary_types():
        print("\n------------- WIP per auxiliary -------------\n")
        print(post_processor.df_aggregated_auxiliary_WIP)

    print("\n------------- Throughput time -------------\n")
    print(post_processor.df_aggregated_throughput_time)

    print("\n------------- Resource states -------------\n")

    print(
        post_processor.df_aggregated_resource_states.copy().set_index(
            ["Resource", "Time_type"]
        )
    )



from __future__ import annotations

from typing import List, Optional, TYPE_CHECKING

from pydantic import BaseModel, TypeAdapter

from prodsys.factories import time_model_factory
from prodsys.models import processes_data

if TYPE_CHECKING:
    from prodsys.adapters import adapter
    from prodsys.simulation import process


class ProcessFactory(BaseModel):
    """
    Factory class that creates and stores `prodsys.simulation` process objects based on the given process data according to `prodsys.models.processes_data.PROCESS_UNION`.

    Args:
        time_model_factory (time_model_factory.TimeModelFactory): Factory that creates time model objects.
        processes (List[process.PROCESS_UNION], optional): List of process objects. Defaults to [] and is filled by the `create_processes` method.
    """

    time_model_factory: time_model_factory.TimeModelFactory
    processes: List[process.PROCESS_UNION] = []

    def create_processes(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates process objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the process data.
        """
        for process_data in adapter.process_data:
            self.add_processes(process_data, adapter)

    def add_processes(
        self,
        process_data: processes_data.PROCESS_DATA_UNION,
        adapter: adapter.ProductionSystemAdapter,
    ):
        values = {"process_data": process_data}
        if not (
            isinstance(process_data, processes_data.CompoundProcessData)
            or isinstance(process_data, processes_data.RequiredCapabilityProcessData)
        ):
            time_model = self.time_model_factory.get_time_model(
                process_data.time_model_id
            )
            values.update({"time_model": time_model})
        else:
            values.update({"time_model": None})
        if "failure_rate" in process_data:
            # TODO: fix this in simulation process to use parameter of process data
            values.update({"failure_rate": process_data.failure_rate})
        if isinstance(process_data, processes_data.TransportProcessData):
            if process_data.loading_time_model_id is not None:
                values.update(
                    {
                        "loading_time_model": self.time_model_factory.get_time_model(
                            process_data.loading_time_model_id
                        )
                    }
                )
            if process_data.unloading_time_model_id is not None:
                values.update(
                    {
                        "unloading_time_model": self.time_model_factory.get_time_model(
                            process_data.unloading_time_model_id
                        )
                    }
                )
        if isinstance(process_data, processes_data.CompoundProcessData):
            contained_processes_data = [
                other_process_data
                for other_process_data in adapter.process_data
                if other_process_data.ID in process_data.process_ids
            ]
            values.update({"contained_processes_data": contained_processes_data})
        if isinstance(process_data, processes_data.LinkTransportProcessData):
            values.update({"links": [[]]})
            self.processes.append(
                TypeAdapter(process.LinkTransportProcess).validate_python(values)
            )
        elif isinstance(process_data, processes_data.ReworkProcessData):
            # TODO: think about getting here the processes and not only ids...
            values.update({"reworked_process_ids": process_data.reworked_process_ids})
            # TODO: fix this in simulation process to use parameter of process data
            values.update({"blocking": process_data.blocking})
            self.processes.append(
                TypeAdapter(process.ReworkProcess).validate_python(values)
            )
        else:
            self.processes.append(
                TypeAdapter(process.PROCESS_UNION).validate_python(values)
            )

    def get_processes_in_order(self, IDs: List[str]) -> List[process.PROCESS_UNION]:
        """
        Returns a list of process objects in the order of the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to sort the process objects.

        Returns:
            List[process.PROCESS_UNION]: List of process objects in the order of the given IDs.
        """
        processes = []
        for ID in IDs:
            for _process in self.processes:
                if _process.process_data.ID == ID:
                    processes.append(_process)

        return processes

    def get_process(self, ID: str) -> Optional[process.PROCESS_UNION]:
        """
        Returns a process object based on the given ID.

        Args:
            ID (str): ID of the process object.

        Raises:
            ValueError: If the process object is not found.

        Returns:
            Optional[process.PROCESS_UNION]: Process object based on the given ID.
        """
        pr = [pr for pr in self.processes if pr.process_data.ID in ID]
        if not pr:
            raise ValueError(f"Process with ID {ID} not found")
        return pr.pop()


from prodsys.simulation import process
